from datetime import datetime
from unittest import TestCase
from unittest import mock

from google.protobuf.timestamp_pb2 import Timestamp
from pytz import UTC

from tecton.cli.plan import IntegrationTestSummaries
from tecton.cli.plan import PlanListItem
from tecton.cli.plan import PlanSummary
from tecton.cli.plan import get_plans_list_items
from tecton_proto.auth import principal__client_pb2 as principal_pb2
from tecton_proto.data import state_update__client_pb2 as state_update_pb2
from tecton_proto.data.state_update__client_pb2 import IntegrationTestJobStatus
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


def make_principle_basic(first_name="john", last_name="smith"):
    return principal_pb2.PrincipalBasic(
        user=principal_pb2.UserBasic(
            okta_id="okta-id",
            first_name=first_name,
            last_name=last_name,
            login_email=f"{first_name}.{last_name}@tecon.ai",
        )
    )


class BaseTestCase(TestCase):
    def mockPatch(self, *args, **kwargs):
        patcher = mock.patch(*args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def setUp(self) -> None:
        self.mock_metadata_service = mock.MagicMock()
        self.mockPatch("tecton._internals.metadata_service.instance", return_value=self.mock_metadata_service)


class TestIntegrationTestSummaries(BaseTestCase):
    def test_from_proto(self):
        resp = IntegrationTestSummaries.from_protobuf(state_update_pb2.SuccessfulPlanOutput())
        self.assertEqual(resp, IntegrationTestSummaries({}))

        resp = IntegrationTestSummaries.from_protobuf(
            state_update_pb2.SuccessfulPlanOutput(
                test_summaries=[state_update_pb2.PlanIntegrationTestSummary(feature_view_name="fv1")]
            )
        )
        self.assertEqual(resp, IntegrationTestSummaries({"fv1": []}))

    def test_summarize_status(self):
        no_tests = []
        only_succeeded = [IntegrationTestJobStatus.JOB_STATUS_SUCCEED, IntegrationTestJobStatus.JOB_STATUS_SUCCEED]
        all_statuses = [
            IntegrationTestJobStatus.JOB_STATUS_NOT_STARTED,
            IntegrationTestJobStatus.JOB_STATUS_RUNNING,
            IntegrationTestJobStatus.JOB_STATUS_CANCELLED,
            IntegrationTestJobStatus.JOB_STATUS_SUCCEED,
            IntegrationTestJobStatus.JOB_STATUS_FAILED,
        ]

        no_failed = [
            IntegrationTestJobStatus.JOB_STATUS_NOT_STARTED,
            IntegrationTestJobStatus.JOB_STATUS_RUNNING,
            IntegrationTestJobStatus.JOB_STATUS_CANCELLED,
            IntegrationTestJobStatus.JOB_STATUS_SUCCEED,
        ]

        running = [
            IntegrationTestJobStatus.JOB_STATUS_NOT_STARTED,
            IntegrationTestJobStatus.JOB_STATUS_RUNNING,
            IntegrationTestJobStatus.JOB_STATUS_SUCCEED,
        ]
        not_started = [
            IntegrationTestJobStatus.JOB_STATUS_NOT_STARTED,
            IntegrationTestJobStatus.JOB_STATUS_SUCCEED,
        ]
        case_list = [
            ("No Tests", no_tests),
            ("Succeeded", only_succeeded),
            ("Failed", all_statuses),
            ("Canceled", no_failed),
            ("Running", running),
            ("Not Started", not_started),
        ]

        for case in case_list:
            with self.subTest(case):
                self.assertEqual(case[0], IntegrationTestSummaries._summarize_status(case[1]))

    def test_summarize_status_for_all_tests(self):
        case_list = [
            {"statuses": {}, "result": "No Tests"},
            {"statuses": {"fv1": [], "fv2": []}, "result": "No Tests"},
            {
                "statuses": {
                    "fv1": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv2": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED, IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv3": [],
                },
                "result": "Succeeded",
            },
            {
                "statuses": {
                    "fv1": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv2": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED, IntegrationTestJobStatus.JOB_STATUS_FAILED],
                },
                "result": "Failed",
            },
        ]
        for case in case_list:
            with self.subTest(case):
                self.assertEqual(
                    case["result"], IntegrationTestSummaries(case["statuses"]).summarize_status_for_all_tests()
                )

    def test_summarize_status_by_fv(self):
        case_list = [
            {"statuses": {}, "result": {}},
            {"statuses": {"fv1": [], "fv2": []}, "result": {"fv1": "No Tests", "fv2": "No Tests"}},
            {
                "statuses": {
                    "fv1": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv2": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED, IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv3": [],
                },
                "result": {"fv1": "Succeeded", "fv2": "Succeeded", "fv3": "No Tests"},
            },
            {
                "statuses": {
                    "fv1": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED],
                    "fv2": [IntegrationTestJobStatus.JOB_STATUS_SUCCEED, IntegrationTestJobStatus.JOB_STATUS_FAILED],
                },
                "result": {"fv1": "Succeeded", "fv2": "Failed"},
            },
        ]
        for case in case_list:
            with self.subTest(case):
                self.assertEqual(case["result"], IntegrationTestSummaries(case["statuses"]).summarize_status_by_fv())


class TestListPlan(BaseTestCase):
    maxDiff = 10000

    def test_plan_list_item_from_proto_no_test_status(self):
        pb_test_date_1 = self.make_fake_date(1)
        pb_test_date_2 = self.make_fake_date(2)
        plan_list_item = PlanListItem.from_proto(
            state_update_pb2.StateUpdateEntry(
                commit_id="fake_commit",
                applied_at=pb_test_date_1,
                applied_by="me",
                created_at=pb_test_date_2,
                created_by="you",
                applied_by_principal=make_principle_basic(),
                workspace="here",
                sdk_version="0.0.0",
                successful_plan_output=state_update_pb2.SuccessfulPlanOutput(
                    test_summaries=[state_update_pb2.PlanIntegrationTestSummary(feature_view_name="fv1")]
                ),
            )
        )
        self.assertEqual(plan_list_item.integration_test_statuses, IntegrationTestSummaries({"fv1": []}))
        self.assertEqual(
            plan_list_item,
            PlanListItem(
                plan_id="fake_commit",
                applied_by="john.smith@tecon.ai(User Email)",
                applied_at=datetime(year=2024, month=1, day=1, tzinfo=UTC),
                created_by="you",
                created_at=datetime(year=2024, month=1, day=2, tzinfo=UTC),
                workspace="here",
                sdk_version="0.0.0",
                integration_test_statuses=IntegrationTestSummaries({"fv1": []}),
            ),
        )

    def test_plan_list_item_from_proto_with_test_status(self):
        pb_test_date_1 = self.make_fake_date(1)
        pb_test_date_2 = self.make_fake_date(2)
        plan_list_item = PlanListItem.from_proto(
            state_update_pb2.StateUpdateEntry(
                commit_id="fake_commit",
                applied_at=pb_test_date_1,
                applied_by="me",
                created_by="you",
                created_at=pb_test_date_2,
                applied_by_principal=make_principle_basic(),
                workspace="here",
                sdk_version="0.0.0",
                successful_plan_output=state_update_pb2.SuccessfulPlanOutput(
                    test_summaries=[
                        state_update_pb2.PlanIntegrationTestSummary(
                            feature_view_name="fv1",
                            job_summaries=[
                                state_update_pb2.IntegrationTestJobSummary(
                                    status=state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_FAILED
                                )
                            ],
                        )
                    ]
                ),
            )
        )
        self.assertEqual(
            plan_list_item,
            PlanListItem(
                plan_id="fake_commit",
                applied_by="john.smith@tecon.ai(User Email)",
                applied_at=datetime(year=2024, month=1, day=1, tzinfo=UTC),
                created_by="you",
                created_at=datetime(year=2024, month=1, day=2, tzinfo=UTC),
                workspace="here",
                sdk_version="0.0.0",
                integration_test_statuses=IntegrationTestSummaries(
                    {"fv1": [state_update_pb2.IntegrationTestJobStatus.JOB_STATUS_FAILED]}
                ),
            ),
        )

    def make_fake_date(self, day):
        test_date = datetime(year=2024, month=1, day=day)
        pb_test_date = Timestamp()
        pb_test_date.FromDatetime(test_date)
        return pb_test_date

    def test_plan_list_items_from_proto_empty(self):
        plan_list_item = PlanListItem.from_proto(state_update_pb2.StateUpdateEntry())
        self.assertEqual(
            plan_list_item,
            PlanListItem(
                plan_id="",
                applied_by="",
                applied_at=None,
                created_by="",
                created_at=None,
                workspace="prod",
                sdk_version="",
                integration_test_statuses=IntegrationTestSummaries({}),
            ),
        )

    def test_get_plans_list_items(self):
        self.mock_metadata_service.GetStateUpdatePlanList.return_value = (
            metadata_service_pb2.GetStateUpdatePlanListResponse(
                entries=[
                    state_update_pb2.StateUpdateEntry(
                        commit_id="fake-commit",
                        applied_by="me",
                        created_by="you",
                        workspace="my_workspace",
                        sdk_version="0.0.0",
                        successful_plan_output=state_update_pb2.SuccessfulPlanOutput(),
                    ),
                    state_update_pb2.StateUpdateEntry(
                        commit_id="fake-commit2",
                        applied_by="you",
                        created_by="you",
                        workspace="my_workspace",
                        sdk_version="0.0.1",
                        successful_plan_output=state_update_pb2.SuccessfulPlanOutput(),
                    ),
                ]
            )
        )
        items = get_plans_list_items("my_workspace", 100)
        self.mock_metadata_service.GetStateUpdatePlanList.assert_called_with(
            metadata_service_pb2.GetStateUpdatePlanListRequest(workspace="my_workspace", limit=100)
        )
        assert items == [
            PlanListItem(
                plan_id="fake-commit",
                applied_by="me",
                applied_at=None,
                created_by="you",
                created_at=None,
                workspace="my_workspace",
                sdk_version="0.0.0",
                integration_test_statuses=IntegrationTestSummaries({}),
            ),
            PlanListItem(
                plan_id="fake-commit2",
                applied_by="you",
                applied_at=None,
                created_by="you",
                created_at=None,
                workspace="my_workspace",
                sdk_version="0.0.1",
                integration_test_statuses=IntegrationTestSummaries({}),
            ),
        ]

    def test_get_plans_list_items_empty(self):
        self.mock_metadata_service.GetStateUpdatePlanList.return_value = (
            metadata_service_pb2.GetStateUpdatePlanListResponse()
        )
        items = get_plans_list_items("my_workspace", 100)
        self.mock_metadata_service.GetStateUpdatePlanList.assert_called_with(
            metadata_service_pb2.GetStateUpdatePlanListRequest(workspace="my_workspace", limit=100)
        )
        assert items == []


class TestGetPlan(BaseTestCase):
    def test_plan_summary_from_proto_empty(self):
        plan = PlanSummary.from_proto(metadata_service_pb2.QueryStateUpdateResponseV2())
        self.assertEqual(
            plan,
            PlanSummary(
                applied=False,
                applied_by="",
                applied_at=None,
                created_by="",
                created_at=None,
                workspace="prod",
                sdk_version="",
                integration_test_statuses=IntegrationTestSummaries(statuses={}),
                plan_url="",
            ),
        )



from __future__ import annotations

from collections import abc
from dataclasses import dataclass
from logging import getLogger
from types import MappingProxyType
from typing import Any
from typing import Callable
from typing import List
from typing import Mapping
from typing import Optional

import numpy
import pyarrow
import torch

from tecton_core import schema
from tecton_core.embeddings import execution_utils
from tecton_core.embeddings.model_container import CustomModelContainerBase
from tecton_core.embeddings.model_container import tecton_type_to_pyarrow
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError


logger = getLogger(__name__)

_PRE_PROCESSOR_FN_NAME = "preprocessor"
_POST_PROCESSOR_FN_NAME = "postprocessor"

_CONTEXT_PARAM = Mapping[str, Any]
_PRE_PROCESSOR_FN = Optional[Callable[[Mapping[str, numpy.ndarray], _CONTEXT_PARAM], Mapping[str, torch.Tensor]]]
_POST_PROCESSOR_FN = Optional[Callable[[torch.Tensor, _CONTEXT_PARAM], numpy.ndarray]]

_MAX_BISECT_RECURSION_LEVEL = 8


def _default_pytorch_preprocessor(input: Mapping[str, numpy.ndarray]) -> Mapping[str, torch.Tensor]:
    return {name: torch.tensor(array) for name, array in input.items()}


class PytorchModelContainer(CustomModelContainerBase):
    _preprocessor: _PRE_PROCESSOR_FN = None
    _postprocessor: _POST_PROCESSOR_FN = None

    def load(self):
        model_module = super().load()
        self._preprocessor = getattr(model_module, _PRE_PROCESSOR_FN_NAME, None)
        self._postprocessor = getattr(model_module, _POST_PROCESSOR_FN_NAME, None)

    def predict(self, input: pyarrow.RecordBatch) -> numpy.ndarray:
        numpy_dict = {
            name: column.to_numpy(zero_copy_only=False) for name, column in zip(input.schema.names, input.columns)
        }
        if self._preprocessor:
            input = self._preprocessor(numpy_dict, self._context)
        else:
            input = _default_pytorch_preprocessor(numpy_dict)

        if not isinstance(input, abc.Mapping):
            msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]', but received '{type(input)}'"
            raise TypeError(msg)

        for k, v in input.items():
            if not isinstance(v, torch.Tensor):
                msg = f"`preprocessor` needs to return a 'Mapping[str, torch.Tensor]' but detected '{k}' returned by preprocessor is an instance of type {type(v)}."
                raise TypeError(msg)

        with torch.no_grad():
            output = self._model(**input)

        if isinstance(output, torch.Tensor):
            output = output.cpu()

        if self._postprocessor:
            output = self._postprocessor(output, self._context)

            if not isinstance(output, numpy.ndarray):
                msg = (
                    f"`postprocessor` needs to return a 'numpy.ndarray' but detected an instance of type {type(output)}"
                )
                raise TypeError(msg)
        else:
            output = output.numpy()

        return output


def pytorch_model_inference(
    batch: pyarrow.RecordBatch,
    *,
    model: PytorchModelContainer,
    input_columns: List[schema.Column],
    output_column: schema.Column,
    cuda_device: str,
    model_input_schema: List[schema.Column],
) -> pyarrow.RecordBatch:
    with torch.device(cuda_device):
        return _pytorch_model_inference(batch, model, input_columns, output_column, model_input_schema)


def _pytorch_model_inference(
    batch: pyarrow.RecordBatch,
    model: PytorchModelContainer,
    input_columns: List[schema.Column],
    output_column: schema.Column,
    model_input_schema: List[schema.Column],
) -> pyarrow.RecordBatch:
    input_batch = batch.select([col.name for col in input_columns])
    model_input_names = [col.name for col in model_input_schema]

    # Rename input columns to model input schema column names.
    # TODO(jiadong): we can use pyarrow built-in `rename_columns` when we upgrade pyarrow to 16.0+.
    input_columns = [input_batch.column(i) for i in range(input_batch.num_columns)]
    input_batch = pyarrow.RecordBatch.from_arrays(input_columns, model_input_names)

    output = _pytorch_model_predict(model=model, input_batch=input_batch)
    output = _convert_pytorch_to_arrow(output, batch.num_rows, tecton_type_to_pyarrow(output_column.dtype))

    batch_with_results = pyarrow.RecordBatch.from_arrays(
        [*batch.columns, output],
        schema=batch.schema.append(pyarrow.field(output_column.name, output.type)),
    )
    return batch_with_results


def _pytorch_model_predict(
    model: PytorchModelContainer, input_batch: pyarrow.RecordBatch, recursion_level: int = 0
) -> numpy.ndarray:
    try:
        return model.predict(input_batch)
    except torch.cuda.OutOfMemoryError as oom_exc:
        logger.warning(torch.cuda.memory_summary())
        logger.warning(f"Out of Memory. {oom_exc}\n")
        execution_utils.gc_all(torch.cuda.current_device())
    return _pytorch_model_predict_with_bisect(model=model, input_batch=input_batch, recursion_level=recursion_level + 1)


def _pytorch_model_predict_with_bisect(
    model: PytorchModelContainer, input_batch: pyarrow.RecordBatch, recursion_level: int
) -> numpy.ndarray:
    if recursion_level == _MAX_BISECT_RECURSION_LEVEL:
        msg = "Inference is out of memory and out of retries."
        raise TectonInternalError(msg)

    logger.warning(
        f"Attempt {recursion_level}/{_MAX_BISECT_RECURSION_LEVEL} to bisect the input batch and trying again."
    )
    input_size = input_batch.num_rows
    mid_point = input_size // 2

    first_half = input_batch.slice(offset=0, length=mid_point)
    second_half = input_batch.slice(offset=mid_point)
    return numpy.concatenate(
        [
            _pytorch_model_predict(model, first_half, recursion_level),
            _pytorch_model_predict(model, second_half, recursion_level),
        ]
    )


def _convert_pytorch_to_arrow(
    output: numpy.ndarray, num_of_rows: int, output_column_type: pyarrow.DataType
) -> pyarrow.Array:
    # If the output is scalar but is 2d (n x 1), then conversion to pyArrow will fail with scalar output_column_type
    # so we need to remove that extra dimension
    if output.ndim == 2 and not isinstance(output_column_type, pyarrow.ListType):
        output = output.squeeze()

    if output.shape[0] != num_of_rows:
        msg = "Output of model should have a single row per input row. "
        raise TectonValidationError(msg)
    return pyarrow.array(list(output), type=output_column_type)


@dataclass
class PytorchModelInferenceFuncConfig:
    _model_dir: str
    _model_file_path: str
    _extra_kwargs: Mapping[str, Any]
    _cuda_device: str

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(
        cls,
        model_dir: str,
        model_file_path: str,
        cuda_device: str,
        input_columns: List[schema.Column],
        output_column: schema.Column,
        model_input_schema: List[schema.Column],
    ) -> PytorchModelInferenceFuncConfig:
        return cls(
            _model_dir=model_dir,
            _model_file_path=model_file_path,
            _cuda_device=cuda_device,
            _extra_kwargs=MappingProxyType(
                {
                    "cuda_device": cuda_device,
                    "input_columns": input_columns,
                    "output_column": output_column,
                    "model_input_schema": model_input_schema,
                }
            ),
        )

    def load(self):
        model_container = PytorchModelContainer(data_dir=self._model_dir, model_file_path=self._model_file_path)
        with torch.device(self._cuda_device):
            model_container.load()

        self._final_kwargs = MappingProxyType(dict(model=model_container, **self._extra_kwargs))
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/transformation__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import basic_info__client_pb2 as tecton__proto_dot_args_dot_basic__info____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.common import framework_version__client_pb2 as tecton__proto_dot_common_dot_framework__version____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.tecton_proto/args/transformation__client.proto\x12\x11tecton_proto.args\x1a*tecton_proto/args/basic_info__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a\x33tecton_proto/common/framework_version__client.proto\x1a$tecton_proto/common/id__client.proto\"\x93\x05\n\x12TransformationArgs\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x46\n\x07version\x18\x07 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x05\x92M\x02\x08\x05R\x07version\x12.\n\x0fprevent_destroy\x18\x08 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x0epreventDestroy\x12S\n\x07options\x18\t \x03(\x0b\x32\x32.tecton_proto.args.TransformationArgs.OptionsEntryB\x05\x92M\x02\x08\x01R\x07options\x12]\n\x13transformation_mode\x18\x03 \x01(\x0e\x32%.tecton_proto.args.TransformationModeB\x05\x92M\x02\x08\x06R\x12transformationMode\x12R\n\ruser_function\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionB\x05\x92M\x02\x08\x06R\x0cuserFunction\x12#\n\tdocstring\x18\x06 \x01(\tB\x05\x92M\x02\x08\x01R\tdocstring\x12\x1d\n\nis_builtin\x18\x05 \x01(\x08R\tisBuiltin\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01*\xdd\x02\n\x12TransformationMode\x12#\n\x1fTRANSFORMATION_MODE_UNSPECIFIED\x10\x00\x12\x1f\n\x1bTRANSFORMATION_MODE_PYSPARK\x10\x01\x12!\n\x1dTRANSFORMATION_MODE_SPARK_SQL\x10\x02\x12\x1e\n\x1aTRANSFORMATION_MODE_PANDAS\x10\x03\x12%\n!TRANSFORMATION_MODE_SNOWFLAKE_SQL\x10\x04\x12\x1e\n\x1aTRANSFORMATION_MODE_PYTHON\x10\x05\x12 \n\x1cTRANSFORMATION_MODE_SNOWPARK\x10\x06\x12$\n TRANSFORMATION_MODE_BIGQUERY_SQL\x10\x08\x12)\n%TRANSFORMATION_MODE_NO_TRANSFORMATION\x10\t\"\x04\x08\x07\x10\x07\x42;\n\x0f\x63om.tecton.argsP\x01Z&github.com/tecton-ai/tecton_proto/args')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.transformation__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001Z&github.com/tecton-ai/tecton_proto/args'
  _TRANSFORMATIONARGS_OPTIONSENTRY._options = None
  _TRANSFORMATIONARGS_OPTIONSENTRY._serialized_options = b'8\001'
  _TRANSFORMATIONARGS.fields_by_name['info']._options = None
  _TRANSFORMATIONARGS.fields_by_name['info']._serialized_options = b'\222M\002\020\001'
  _TRANSFORMATIONARGS.fields_by_name['version']._options = None
  _TRANSFORMATIONARGS.fields_by_name['version']._serialized_options = b'\222M\002\010\005'
  _TRANSFORMATIONARGS.fields_by_name['prevent_destroy']._options = None
  _TRANSFORMATIONARGS.fields_by_name['prevent_destroy']._serialized_options = b'\222M\002\010\001'
  _TRANSFORMATIONARGS.fields_by_name['options']._options = None
  _TRANSFORMATIONARGS.fields_by_name['options']._serialized_options = b'\222M\002\010\001'
  _TRANSFORMATIONARGS.fields_by_name['transformation_mode']._options = None
  _TRANSFORMATIONARGS.fields_by_name['transformation_mode']._serialized_options = b'\222M\002\010\006'
  _TRANSFORMATIONARGS.fields_by_name['user_function']._options = None
  _TRANSFORMATIONARGS.fields_by_name['user_function']._serialized_options = b'\222M\002\010\006'
  _TRANSFORMATIONARGS.fields_by_name['docstring']._options = None
  _TRANSFORMATIONARGS.fields_by_name['docstring']._serialized_options = b'\222M\002\010\001'
  _TRANSFORMATIONMODE._serialized_start=968
  _TRANSFORMATIONMODE._serialized_end=1317
  _TRANSFORMATIONARGS._serialized_start=306
  _TRANSFORMATIONARGS._serialized_end=965
  _TRANSFORMATIONARGS_OPTIONSENTRY._serialized_start=907
  _TRANSFORMATIONARGS_OPTIONSENTRY._serialized_end=965
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

CLUSTER_DEFAULT: PythonVersion
DESCRIPTOR: _descriptor.FileDescriptor
PYTHON_3_9_0: PythonVersion
PYTHON_3_9_13: PythonVersion

class PythonVersion(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from __future__ import annotations

from datetime import datetime
from datetime import timedelta
from datetime import timezone
from enum import Enum
from typing import Optional

import attrs

from tecton_core.time_utils import datetime_to_proto
from tecton_core.time_utils import timedelta_to_proto
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2


# Lowest Pandas Timestamp
UNBOUNDED_PAST_TIMESTAMP = datetime(1677, 9, 22).replace(tzinfo=timezone.utc)

# Highest Pandas Timestamp
UNBOUNDED_FUTURE_TIMESTAMP = datetime(2262, 4, 11).replace(tzinfo=timezone.utc)


@attrs.define(frozen=True)
class FilterDateTime:
    """
    Represents a time reference for a FilteredSource. This can be a specific timestamp, or a time offset from a
    TectonTimeConstant.
    """

    time_reference: Optional[TectonTimeConstant] = None
    offset: Optional[timedelta] = timedelta(0)
    timestamp: Optional[datetime] = None

    def to_args_proto(self) -> pipeline_pb2.FilterDateTime:
        filter_date_time = pipeline_pb2.FilterDateTime()
        if self.time_reference:
            relative_time = pipeline_pb2.RelativeTime(
                time_reference=self.time_reference.to_args_proto(), offset=timedelta_to_proto(self.offset)
            )
            filter_date_time.relative_time.CopyFrom(relative_time)
        elif self.timestamp:
            filter_date_time.timestamp.CopyFrom(datetime_to_proto(self.timestamp))

        return filter_date_time

    @classmethod
    def from_proto(cls, proto: pipeline_pb2.FilterDateTime) -> FilterDateTime:
        if proto.HasField("relative_time"):
            time_reference = TectonTimeConstant.from_proto(proto.relative_time.time_reference)
            offset = proto.relative_time.offset.ToTimedelta()
            return FilterDateTime(time_reference=time_reference, offset=offset)
        elif proto.HasField("timestamp"):
            return FilterDateTime(timestamp=proto.timestamp.ToDatetime())
        else:
            msg = "FilterDateTime proto must have either relative_time or timestamp"
            raise ValueError(msg)

    def is_materialization_limit(self) -> bool:
        return self.time_reference is not None and self.time_reference.is_materialization_limit()

    def is_unbounded_limit(self) -> bool:
        return self.time_reference is not None and self.time_reference.is_unbounded_limit()

    def to_datetime(self, exact_reference: Optional[datetime] = None) -> Optional[datetime]:
        """
        Convert the FilterDateTime to a datetime object. Exact reference should be the
        materialization start or end time of the Feature View (if applicable).

        :param exact_reference: The Materialization start or end time of the Feature View which
        will be used when MATERIALIZATION_START_TIME or MATERIALIZATION_END_TIME is used as the time_reference.
        :return: A datetime object representing the FilterDateTime.
        """
        if self.time_reference:
            if (
                self.time_reference == TectonTimeConstant.MATERIALIZATION_START_TIME
                or self.time_reference == TectonTimeConstant.MATERIALIZATION_END_TIME
            ):
                if exact_reference is None:
                    return None
                dt = exact_reference + self.offset
            elif self.time_reference == TectonTimeConstant.UNBOUNDED_PAST:
                # Close to `datetime.min` but can still be represented as a timestamp
                dt = UNBOUNDED_PAST_TIMESTAMP
            elif self.time_reference == TectonTimeConstant.UNBOUNDED_FUTURE:
                dt = UNBOUNDED_FUTURE_TIMESTAMP
            else:
                err = f"Unsupported TectonTimeConstant: {self.time_reference}"
                raise ValueError(err)
        elif self.timestamp:
            dt = self.timestamp
        else:
            err = "FilterDateTime must have a time_reference or timestamp"
            raise ValueError(err)

        return dt.replace(tzinfo=timezone.utc)


class TectonTimeConstant(Enum):
    """
    TectonTimeConstant is an enumeration of time constants used in Tecton that could be useful in defining time-based
    filters in Feature Views. These constants can be used in conjunction with timedelta to define time ranges for
    filters using DataSource.select_range().

    Example:
    ```
    @batch_feature_view(
        ...
        sources=[
            transactions_source.select_range(
                start_time=TectonTimeConstant.MATERIALIZATION_START_TIME - timedelta(days=7),
                end_time=TectonTimeConstant.MATERIALIZATION_END_TIME
            )
        ]
        ...
    )
    def my_feature_view(transactions_source):
        ...
    ```
    """

    MATERIALIZATION_START_TIME = 1
    MATERIALIZATION_END_TIME = 2
    UNBOUNDED_PAST = 3
    UNBOUNDED_FUTURE = 4

    def __add__(self, other: timedelta) -> FilterDateTime:
        if not isinstance(other, timedelta):
            err = "Can only add `timedelta` objects to a TectonTimeConstant"
            raise TypeError(err)
        if self == TectonTimeConstant.UNBOUNDED_PAST or self == TectonTimeConstant.UNBOUNDED_FUTURE:
            err = f"Cannot add a timedelta to {self}"
            raise ValueError(err)
        return FilterDateTime(time_reference=self, offset=other)

    def __sub__(self, other: timedelta) -> FilterDateTime:
        if not isinstance(other, timedelta):
            err = "Can only subtract `timedelta` objects from a TectonTimeConstant"
            raise TypeError(err)
        if self == TectonTimeConstant.UNBOUNDED_PAST or self == TectonTimeConstant.UNBOUNDED_FUTURE:
            err = f"Cannot add a timedelta offset to {self}"
            raise ValueError(err)
        return FilterDateTime(time_reference=self, offset=-other)

    def is_materialization_limit(self) -> bool:
        return self in [
            TectonTimeConstant.MATERIALIZATION_START_TIME,
            TectonTimeConstant.MATERIALIZATION_END_TIME,
        ]

    def is_unbounded_limit(self) -> bool:
        return self in [
            TectonTimeConstant.UNBOUNDED_PAST,
            TectonTimeConstant.UNBOUNDED_FUTURE,
        ]

    def to_args_proto(self) -> pipeline_pb2.TimeReference:
        if self == TectonTimeConstant.MATERIALIZATION_START_TIME:
            return pipeline_pb2.TimeReference.TIME_REFERENCE_MATERIALIZATION_START_TIME
        elif self == TectonTimeConstant.MATERIALIZATION_END_TIME:
            return pipeline_pb2.TimeReference.TIME_REFERENCE_MATERIALIZATION_END_TIME
        elif self == TectonTimeConstant.UNBOUNDED_PAST:
            return pipeline_pb2.TimeReference.TIME_REFERENCE_UNBOUNDED_PAST
        elif self == TectonTimeConstant.UNBOUNDED_FUTURE:
            return pipeline_pb2.TimeReference.TIME_REFERENCE_UNBOUNDED_FUTURE
        else:
            err = f"Unsupported TectonTimeConstant: {self}"
            raise ValueError(err)

    @classmethod
    def from_proto(cls, proto: pipeline_pb2.TimeReference) -> Optional[TectonTimeConstant]:
        if proto == pipeline_pb2.TimeReference.TIME_REFERENCE_MATERIALIZATION_START_TIME:
            return TectonTimeConstant.MATERIALIZATION_START_TIME
        elif proto == pipeline_pb2.TimeReference.TIME_REFERENCE_MATERIALIZATION_END_TIME:
            return TectonTimeConstant.MATERIALIZATION_END_TIME
        elif proto == pipeline_pb2.TimeReference.TIME_REFERENCE_UNBOUNDED_PAST:
            return TectonTimeConstant.UNBOUNDED_PAST
        elif proto == pipeline_pb2.TimeReference.TIME_REFERENCE_UNBOUNDED_FUTURE:
            return TectonTimeConstant.UNBOUNDED_FUTURE
        else:
            return None



import json
import logging
import multiprocessing
import os
import uuid
from pathlib import Path
from typing import List
from typing import Union

import attrs
import boto3
import pyarrow
import pyarrow.dataset
import pyarrow.fs
import pyarrow.parquet as pq
import ray
from pyarrow._dataset import WrittenFile

from tecton_core import conf
from tecton_core import duckdb_factory
from tecton_core import offline_store
from tecton_core.arrow import PARQUET_WRITE_OPTIONS
from tecton_core.compute_mode import ComputeMode
from tecton_core.data_processing_utils import pyarrow_split_spine
from tecton_core.duckdb_factory import DuckDBConfig
from tecton_core.offline_store import DATASET_PARTITION_SIZE
from tecton_core.query.dialect import Dialect
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import PyarrowDataframeWrapper
from tecton_core.query.nodes import StagingNode
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_core.query.retrieval_params import GetFeaturesForEventsParams
from tecton_core.query.retrieval_params import GetFeaturesInRangeParams
from tecton_core.query.rewrite import rewrite_tree_for_spine
from tecton_core.query_consts import valid_to
from tecton_materialization.common.dataset_generation import get_features_from_params
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.common.task_params import feature_definition_from_feature_service_task_params
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.common.task_params import get_features_params_from_task_params
from tecton_materialization.ray.delta import DeltaWriter
from tecton_materialization.ray.delta import OfflineStoreParams
from tecton_materialization.ray.delta import write
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.job_status import MonitoringContextProvider
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.materialization_utils import get_secret_resolver
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_materialization.ray.nodes import TimeSpec
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.materialization_task__client_pb2 import DatasetGenerationParameters
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.materialization.params__client_pb2 import SecretMaterializationTaskParams
from tecton_proto.offlinestore.delta import transaction_writer__client_pb2 as transaction_writer_pb2


# fewer threads means more memory per thread
DEFAULT_NUM_THREADS = multiprocessing.cpu_count() // 2

DEFAULT_SPINE_SPLIT_DIR_STR = "/tmp/spine_split"
# Default output buffer dir for storing output results before uploading to S3 Delta table.
DEFAULT_OUTPUT_BUFFER_DIR_STR = "/tmp/output_buffer"
# Default chunk size for merging and uploading output files to S3 Delta table.
DEFAULT_OUTPUT_CHUNK_SIZE = 30

# With the default setting, a m5.2xlarge instance(8 cpu, 32G memory) can run 4 subtasks in parallel.
# Worth noting that by default only 70% of memory (32G*0.7=~22G) can be used by Ray subtasks.
DEFAULT_RAY_THREAD_CPU = 2
DEFAULT_RAY_THREAD_MEMORY_IN_BYTES = 5 * 1024 * 1024 * 1024  # 5G


logger = logging.getLogger(__name__)


def run_dataset_generation(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    # retrieve current region
    conf.set("CLUSTER_REGION", boto3.Session().region_name)
    conf.set("DUCKDB_NTHREADS", DEFAULT_NUM_THREADS)

    assert materialization_task_params.dataset_generation_task_info.HasField("dataset_generation_parameters")
    dataset_generation_params = materialization_task_params.dataset_generation_task_info.dataset_generation_parameters
    params = get_features_params_from_task_params(materialization_task_params, compute_mode=ComputeMode.RIFT)

    config_overrides = json.loads(dataset_generation_params.extra_config.get("tecton_config", "{}"))
    for k, v in config_overrides.items():
        conf.set(k, v)

    store_params = OfflineStoreParams(
        feature_view_id=params.fco.id,
        feature_view_name=params.fco.name,
        schema=dataset_generation_params.expected_schema,
        time_spec=None,
        feature_store_format_version=None,
        batch_schedule=None,
    )

    table = get_delta_writer(
        materialization_task_params,
        store_params=store_params,
        table_uri=dataset_generation_params.result_path,
        join_keys=params.join_keys,
    )

    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Store results to dataset location",
    )

    if should_not_use_ray_threads(dataset_generation_params):
        logger.info("Using one thread for dataset generation")
        run_dataset_generation_with_one_thread(
            materialization_task_params,
            dataset_generation_params,
            job_status_client,
            executor,
            params,
            table,
            offline_stage_monitor,
        )
    else:
        logger.info("Using Ray threads for dataset generation")
        run_dataset_generation_with_ray_tasks(
            materialization_task_params,
            secret_materialization_task_params,
            dataset_generation_params,
            params,
            table,
            offline_stage_monitor,
        )


def should_not_use_ray_threads(dataset_generation_params: DatasetGenerationParameters) -> bool:
    disable_ray_str = str(dataset_generation_params.extra_config.get("disable_ray"))
    return "disable_ray" in dataset_generation_params.extra_config and disable_ray_str.lower() == "true"


def run_dataset_generation_with_ray_tasks(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
    dataset_generation_params: DatasetGenerationParameters,
    params: Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams],
    table: DeltaWriter,
    offline_stage_monitor: MonitoringContextProvider,
):
    if "ray_thread_cpu" not in dataset_generation_params.extra_config:
        cpu_per_task = DEFAULT_RAY_THREAD_CPU
    else:
        cpu_per_task = int(dataset_generation_params.extra_config["ray_thread_cpu"])

    if "ray_thread_memory_in_bytes" not in dataset_generation_params.extra_config:
        memory_per_task_in_bytes = DEFAULT_RAY_THREAD_MEMORY_IN_BYTES
    else:
        memory_per_task_in_bytes = int(dataset_generation_params.extra_config["ray_thread_memory_in_bytes"])

    target_scanned_offline_rows_per_partition = None
    if "target_scanned_offline_rows_per_partition" in dataset_generation_params.extra_config:
        target_scanned_offline_rows_per_partition = int(
            dataset_generation_params.extra_config.get("target_scanned_offline_rows_per_partition")
        )

    tasks = []
    upload_tasks = []
    with offline_stage_monitor():
        if isinstance(params, GetFeaturesForEventsParams):
            time_column = params.timestamp_key
            spine_data = _load_user_provided_data(params.events)
            if conf.get_bool("DUCKDB_ENABLE_SPINE_SPLIT") or target_scanned_offline_rows_per_partition:
                spine_split_tables = pyarrow_split_spine(
                    spine_data,
                    params.join_keys,
                    _extract_fdw_list(materialization_task_params),
                    time_column,
                    target_scanned_offline_rows_per_partition,
                )
                spine_split_files = write_split_tables_to_files(spine_split_tables)
            else:
                spine_split_files = write_split_tables_to_files([spine_data])
            num_tasks = len(spine_split_files)
            for idx, spine_split_file in enumerate(spine_split_files):
                spine_chunk = read_table_from_file(spine_split_file)
                tasks.append(
                    run_one_dataset_generation_ray_task.options(
                        num_cpus=cpu_per_task, memory=memory_per_task_in_bytes
                    ).remote(
                        idx,
                        num_tasks,
                        spine_chunk,
                        time_column,
                        materialization_task_params,
                        secret_materialization_task_params,
                        os.environ.get("TEST_TMPDIR", None),
                        cpu_per_task,
                        memory_per_task_in_bytes,
                    )
                )
        elif isinstance(params, GetFeaturesInRangeParams):
            time_column = valid_to()
            idx = 0
            num_tasks = 1  # Adjust if needed for parallelism
            tasks.append(
                run_one_dataset_generation_ray_task.options(
                    num_cpus=cpu_per_task, memory=memory_per_task_in_bytes
                ).remote(
                    idx,
                    num_tasks,
                    None,  # No spine chunk needed
                    time_column,
                    materialization_task_params,
                    secret_materialization_task_params,
                    os.environ.get("TEST_TMPDIR", None),
                    cpu_per_task,
                    memory_per_task_in_bytes,
                )
            )
        else:
            error = f"Unsupported params type: {type(params)}"
            raise ValueError(error)

        remaining_tasks = tasks
        while remaining_tasks:
            ready_tasks, remaining_tasks = ray.wait(
                remaining_tasks, num_returns=min(DEFAULT_OUTPUT_CHUNK_SIZE, len(remaining_tasks)), timeout=None
            )

            files_list = ray.get(ready_tasks)
            local_files_list = []
            for sublist in files_list:
                local_files_list.extend(sublist)

            upload_tasks.append(
                upload_to_delta_table.options(num_cpus=cpu_per_task, memory=memory_per_task_in_bytes).remote(
                    local_files_list, dataset_generation_params.result_path, params.join_keys
                )
            )

        files_list = ray.get(upload_tasks)

        # Each Ray task returns a list of AddFile objects. We need to commit all of them to the Delta table.
        files = [file for sublist in files_list for file in sublist]
        table.add_files(files)
        table.commit()


@ray.remote
def run_one_dataset_generation_ray_task(
    idx,
    num_tasks,
    spine_chunk,
    time_column,
    materialization_task_params,
    secret_materialization_task_params,
    test_tmpdir,
    cpu_per_task,
    memory_per_task_in_bytes,
) -> List[transaction_writer_pb2.AddFile]:
    conf.set("DUCKDB_DEBUG", "true")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "rift")
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")

    if test_tmpdir:
        duckdb_factory.set_home_dir_override(test_tmpdir)

    pyarrow.set_cpu_count(cpu_per_task)

    job_status_client = JobStatusClient(JobMetadataClient.for_params(materialization_task_params))
    job_status_client.set_query_index(idx, num_tasks)
    secret_resolver = get_secret_resolver(secret_materialization_task_params.secret_service_params)
    executor = QueryTreeExecutor(
        monitor=job_status_client,
        secret_resolver=secret_resolver,
        duckdb_config=DuckDBConfig(
            num_threads=cpu_per_task,
            memory_limit_in_bytes=memory_per_task_in_bytes,
            use_unique_extension_path=True,
        ),
    )

    params = get_features_params_from_task_params(materialization_task_params, compute_mode=ComputeMode.RIFT)
    if isinstance(params, GetFeaturesForEventsParams):
        qt = get_features_from_params(params, spine=PyarrowDataframeWrapper(spine_chunk))
    elif isinstance(params, GetFeaturesInRangeParams):
        entities = PyarrowDataframeWrapper(pq.read_table(params.entities)) if params.entities is not None else None
        qt = get_features_from_params(params, entities=entities)
    else:
        error = f"Unsupported params type: {type(params)}"
        raise ValueError(error)

    qt = _add_partition_column(qt, time_column)
    rewrite_tree_for_spine(qt)
    reader = executor.exec_qt(qt).result_table
    return write_to_tmp_buffer_table(reader)


def write_to_delta_table(
    reader: Union[pyarrow.Table, pyarrow.RecordBatchReader],
    table_uri: str,
    join_keys: List[str],
) -> List[transaction_writer_pb2.AddFile]:
    fs, base_path = pyarrow.fs.FileSystem.from_uri(table_uri)
    partitioning = pyarrow.dataset.partitioning(
        pyarrow.schema([(offline_store.TIME_PARTITION, pyarrow.string())]), flavor="hive"
    )
    return write(reader, base_path, table_uri, join_keys, fs, partitioning)


def _load_user_provided_data(path: str) -> pyarrow.Table:
    spine = pq.read_table(path)
    logger.info(f"Reading spine with shape {spine.shape} and memory usage {spine.nbytes}")
    logger.info(spine.schema)
    return spine


def run_dataset_generation_with_one_thread(
    materialization_task_params: MaterializationTaskParams,
    dataset_generation_params: DatasetGenerationParameters,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
    params: Union[GetFeaturesInRangeParams, GetFeaturesForEventsParams],
    table: DeltaWriter,
    offline_stage_monitor: MonitoringContextProvider,
):
    qts = []

    if isinstance(params, GetFeaturesForEventsParams):
        time_column = params.timestamp_key
        spine_data = _load_user_provided_data(params.events)

        if conf.DUCKDB_ENABLE_SPINE_SPLIT_FF.enabled():
            target_scanned_offline_rows_per_partition = None
            if "target_scanned_offline_rows_per_partition" in dataset_generation_params.extra_config:
                target_scanned_offline_rows_per_partition = int(
                    dataset_generation_params.extra_config.get("target_scanned_offline_rows_per_partition")
                )

            spine_split_files = pyarrow_split_spine(
                spine_data,
                params.join_keys,
                _extract_fdw_list(materialization_task_params),
                time_column,
                # DEFAULT_TARGET_SCANNED_OFFLINE_ROWS_PER_PARTITION will be used.
                target_scanned_offline_rows_per_partition,
            )
            for spine_split_file in spine_split_files:
                spine_chunk = read_table_from_file(spine_split_file)
                qt = get_features_from_params(params, spine=PyarrowDataframeWrapper(spine_chunk))
                qts.append(qt)
        else:
            qts.append(get_features_from_params(params, spine=PyarrowDataframeWrapper(spine_data)))
    elif isinstance(params, GetFeaturesInRangeParams):
        time_column = valid_to()
        entities = (
            PyarrowDataframeWrapper(_load_user_provided_data(params.entities)) if params.entities is not None else None
        )
        qts.append(get_features_from_params(params, entities=entities))
    else:
        error = f"Unsupported params type: {type(params)}"
        raise ValueError(error)

    qts = [_add_partition_column(qt, time_column) for qt in qts]

    for idx, qt in enumerate(qts):
        job_status_client.set_query_index(idx, len(qts))
        rewrite_tree_for_spine(qt)
        reader = executor.exec_qt(qt).result_table

        with offline_stage_monitor():
            table.write(reader)

    table.commit()


def _extract_fdw_list(materialization_task_params: MaterializationTaskParams):
    if materialization_task_params.HasField("feature_view"):
        fdw_list = [feature_definition_from_task_params(materialization_task_params)]
    else:
        fdw_list = feature_definition_from_feature_service_task_params(materialization_task_params)
    return fdw_list


def _add_partition_column(qt: NodeRef, time_column) -> NodeRef:
    """
    Injects AddTimePartitionNode either before StagingNode(step=AGGREGATION) or at the top of the tree.
    The aim is to run this node before ODFV (if it is present) to make it part of DuckDB query.
    """

    def create_node(input_node: NodeRef) -> NodeRef:
        return AddTimePartitionNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=input_node,
            time_spec=TimeSpec(
                timestamp_key=time_column,
                time_column=time_column,
                partition_size=DATASET_PARTITION_SIZE,
                partition_is_anchor=False,
            ),
        ).as_ref()

    def inject(tree: NodeRef) -> bool:
        """
        Traverse over the tree and return True if AddTimePartitionNode was injected before StagingNode(step=AGGREGATION)
        """
        injected = False

        if isinstance(tree.node, StagingNode) and QueryTreeStep.AGGREGATION == tree.node.query_tree_step:
            prev_input = tree.node.input_node
            new_input = create_node(prev_input)
            tree.node = attrs.evolve(tree.node, input_node=new_input)
            injected = True

        return injected or any(inject(tree=i) for i in tree.inputs)

    if inject(qt):
        return qt

    # add node at the top
    return create_node(qt)


def write_split_tables_to_files(split_tables: List[pyarrow.Table]) -> List[Path]:
    # Generate a unique directory name
    split_dir = Path(f"{DEFAULT_SPINE_SPLIT_DIR_STR}_{uuid.uuid4()}")
    split_dir.mkdir(exist_ok=True)

    # Write each split table to a file using Arrow IPC format, which has a better performance than Parquet for Inter Process Communication.
    split_files = []
    for table in split_tables:
        file_path = split_dir / f"split_{uuid.uuid4()}.arrow"
        with pyarrow.ipc.RecordBatchFileWriter(str(file_path), table.schema) as writer:
            writer.write_table(table)
        split_files.append(file_path)

    return split_files


def read_table_from_file(file_path: Path) -> pyarrow.Table:
    if not file_path.exists():
        msg = f"File not found: {file_path}"
        raise FileNotFoundError(msg)

    with pyarrow.ipc.RecordBatchFileReader(str(file_path)) as reader:
        table = reader.read_all()

    return table


def write_to_tmp_buffer_table(
    reader: Union[pyarrow.Table, pyarrow.RecordBatchReader],
) -> List[Path]:
    adds = []
    failed = False

    def local_visit_file(f: WrittenFile):
        try:
            adds.append(Path(f.path))
        except Exception as e:
            # Log and raise exceptions for debugging
            nonlocal failed
            failed = True
            raise e

    pyarrow.dataset.write_dataset(
        data=reader,
        filesystem=pyarrow.fs.LocalFileSystem(),
        base_dir=DEFAULT_OUTPUT_BUFFER_DIR_STR,
        format=pyarrow.dataset.ParquetFileFormat(),
        file_options=PARQUET_WRITE_OPTIONS,
        basename_template=f"{uuid.uuid4()}-part-{{i}}.parquet",
        file_visitor=local_visit_file,
        existing_data_behavior="overwrite_or_ignore",
    )

    if failed:
        msg = "file visitor failed"
        raise Exception(msg)

    return adds


@ray.remote
def upload_to_delta_table(file_list, result_path, join_keys):
    dataset = pyarrow.dataset.dataset(file_list, format=pyarrow.dataset.ParquetFileFormat())
    reader = dataset.scanner().to_reader()

    output_files = write_to_delta_table(reader, result_path, join_keys)

    for file_path in file_list:
        os.remove(file_path)

    return output_files



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DATA_TYPE_ARRAY: DataTypeEnum
DATA_TYPE_BOOL: DataTypeEnum
DATA_TYPE_FLOAT32: DataTypeEnum
DATA_TYPE_FLOAT64: DataTypeEnum
DATA_TYPE_INT32: DataTypeEnum
DATA_TYPE_INT64: DataTypeEnum
DATA_TYPE_MAP: DataTypeEnum
DATA_TYPE_STRING: DataTypeEnum
DATA_TYPE_STRUCT: DataTypeEnum
DATA_TYPE_TIMESTAMP: DataTypeEnum
DATA_TYPE_TIMESTAMP_INT96: DataTypeEnum
DATA_TYPE_UNKNOWN: DataTypeEnum
DESCRIPTOR: _descriptor.FileDescriptor

class DataType(_message.Message):
    __slots__ = ["array_element_type", "map_key_type", "map_value_type", "struct_fields", "type"]
    ARRAY_ELEMENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    MAP_KEY_TYPE_FIELD_NUMBER: _ClassVar[int]
    MAP_VALUE_TYPE_FIELD_NUMBER: _ClassVar[int]
    STRUCT_FIELDS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    array_element_type: DataType
    map_key_type: DataType
    map_value_type: DataType
    struct_fields: _containers.RepeatedCompositeFieldContainer[StructField]
    type: DataTypeEnum
    def __init__(self, type: _Optional[_Union[DataTypeEnum, str]] = ..., array_element_type: _Optional[_Union[DataType, _Mapping]] = ..., struct_fields: _Optional[_Iterable[_Union[StructField, _Mapping]]] = ..., map_key_type: _Optional[_Union[DataType, _Mapping]] = ..., map_value_type: _Optional[_Union[DataType, _Mapping]] = ...) -> None: ...

class StructField(_message.Message):
    __slots__ = ["data_type", "name"]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    data_type: DataType
    name: str
    def __init__(self, name: _Optional[str] = ..., data_type: _Optional[_Union[DataType, _Mapping]] = ...) -> None: ...

class DataTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

AUDIT_LOG_METADATA_FIELD_NUMBER: _ClassVar[int]
DESCRIPTOR: _descriptor.FileDescriptor
OPTIONS_FIELD_NUMBER: _ClassVar[int]
TRANSFORM_LEGACY_ROLES: AuditLogTransform
TRANSFORM_REDACT_STRING: AuditLogTransform
TRANSFORM_UNSPECIFIED: AuditLogTransform
VISIBILITY_OMIT: Visibility
VISIBILITY_UNSPECIFIED: Visibility
VISIBILITY_VISIBLE: Visibility
audit_log_metadata: _descriptor.FieldDescriptor
options: _descriptor.FieldDescriptor

class AuditLogMetadata(_message.Message):
    __slots__ = ["customer_facing_event_name", "version", "write_to_audit_log"]
    CUSTOMER_FACING_EVENT_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    WRITE_TO_AUDIT_LOG_FIELD_NUMBER: _ClassVar[int]
    customer_facing_event_name: str
    version: str
    write_to_audit_log: bool
    def __init__(self, write_to_audit_log: bool = ..., customer_facing_event_name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class AuditLogOptions(_message.Message):
    __slots__ = ["transform", "visibility"]
    TRANSFORM_FIELD_NUMBER: _ClassVar[int]
    VISIBILITY_FIELD_NUMBER: _ClassVar[int]
    transform: AuditLogTransform
    visibility: Visibility
    def __init__(self, visibility: _Optional[_Union[Visibility, str]] = ..., transform: _Optional[_Union[AuditLogTransform, str]] = ...) -> None: ...

class Visibility(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class AuditLogTransform(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import datetime
import itertools
import time
from functools import reduce
from operator import and_
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pandas
import pyarrow

from tecton_core import conf
from tecton_core import duckdb_factory
from tecton_core import id_helper
from tecton_core import specs
from tecton_core.errors import AccessError
from tecton_core.offline_store import BotoOfflineStoreOptionsProvider
from tecton_core.query.errors import SQLCompilationError
from tecton_core.query.errors import UserCodeError
from tecton_core.query.executor_params import ExecutionContext
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.pandas.node import ArrowExecNode
from tecton_core.schema import Schema
from tecton_core.schema_validation import tecton_schema_to_arrow_schema
from tecton_core.specs import DatetimePartitionColumnSpec
from tecton_core.time_utils import get_timezone_aware_datetime
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_proto.data import batch_data_source__client_pb2 as batch_data_source_pb2


@attrs.frozen
class RedshiftDataSourceScanNode(ArrowExecNode):
    ds: specs.DataSourceSpec
    ds_node: Optional[DataSourceNode]
    is_stream: bool = attrs.field()
    start_time: Optional[datetime.datetime] = None
    end_time: Optional[datetime.datetime] = None

    @classmethod
    def from_node_input(cls, query_node: DataSourceScanNode) -> "RedshiftDataSourceScanNode":
        assert isinstance(query_node.ds.batch_source, specs.RedshiftSourceSpec)
        return cls.from_node_inputs(query_node, input_node=NodeRef(query_node))

    def as_str(self):
        return f"RedshiftDataSourceScanNode for {self.ds.name}"

    def to_arrow_reader(self, context: ExecutionContext) -> pyarrow.RecordBatchReader:
        try:
            import redshift_connector
        except ImportError:
            msg = "Please install `redshift-connector` package to connect to Redshift."
            raise ImportError(msg)
        batch_source = self.ds.batch_source
        assert isinstance(batch_source, specs.RedshiftSourceSpec)

        host, port, database = self._parse_redshift_endpoint(batch_source.endpoint)
        # Note: keeping use of conf for secrets for parity with the Spark connector
        # TODO: add secrets support to RedshiftConfig
        user = conf.get_or_none("REDSHIFT_USER")
        password = conf.get_or_none("REDSHIFT_PASSWORD")

        # assuming IAM role is authenticated if pw not setup
        if not password:
            is_iam_auth = True
            # db_user is the user required for IAM role to use; 'awsuser' is the default user.
            db_user = user or "awsuser"
            cluster_identifier = host.split(".")[0]
        else:
            is_iam_auth = False
            db_user = None
            cluster_identifier = None

        chunk_size = 100_000
        execute_sql = self.input_node.to_sql()
        try:
            with redshift_connector.connect(
                host=host,
                port=int(port),
                database=database,
                user=user,
                password=password,
                iam=is_iam_auth,
                db_user=db_user,
                cluster_identifier=cluster_identifier,
            ) as conn:
                with conn.cursor() as cursor:
                    cursor.execute(execute_sql)
                    batches = []
                    while True:
                        df_chunk = cursor.fetch_dataframe(num=chunk_size)
                        # Note: if we stop supporting pandas post_processor we can improve the memory redundancy here
                        if batch_source.post_processor:
                            try:
                                df_chunk = batch_source.post_processor(df_chunk)
                            except Exception as e:
                                msg = f"Post processor function of data source `{self.ds.name}` failed with exception: {e}"
                                raise UserCodeError(msg) from e
                        if df_chunk.empty:
                            break
                        record_batch = pyarrow.RecordBatch.from_pandas(df_chunk, preserve_index=False)
                        batches.append(record_batch)
        except redshift_connector.InterfaceError as ie:
            msg = f"Unable to establish a connection with Redshift.\n{ie}"
            raise AccessError(msg) from ie
        except redshift_connector.ProgrammingError as pe:
            raise SQLCompilationError(str(pe), execute_sql) from pe

        if batches:
            return pyarrow.RecordBatchReader.from_batches(batches[0].schema, batches)
        else:
            # Note: may be neater to infer schema from cursor.descriptor (name, type_code)
            empty_batch = pyarrow.RecordBatch.from_pandas(df_chunk, preserve_index=False)
            return pyarrow.RecordBatchReader.from_batches(empty_batch.schema, [empty_batch])

    def _parse_redshift_endpoint(self, endpoint: str) -> Tuple[str, str, str]:
        """
        split endpoint in to separate host, port, database fields
        Example:
        'cluster-name.abc123xyz.us-west-2.redshift.amazonaws.com:5439/my_db'
        cluster-name, 5439, my_db
        """
        host_and_port, database = endpoint.rsplit("/", 1)
        host, port = host_and_port.rsplit(":", 1)
        return host, port, database


@attrs.define
class FileDataSourceScanNode(ArrowExecNode):
    ds: specs.DataSourceSpec
    ds_node: Optional[DataSourceNode]
    is_stream: bool = attrs.field()
    start_time: Optional[datetime.datetime] = None
    end_time: Optional[datetime.datetime] = None

    @classmethod
    def from_node_input(cls, query_node: DataSourceScanNode) -> "FileDataSourceScanNode":
        assert isinstance(query_node.ds.batch_source, specs.FileSourceSpec)
        return cls.from_node_inputs(query_node, input_node=None)

    @property
    def spec(self) -> specs.FileSourceSpec:
        return self.ds.batch_source

    def to_arrow_reader(self, context: ExecutionContext) -> pyarrow.RecordBatchReader:
        import duckdb

        file_uri = self.spec.uri
        timestamp_field = self.spec.timestamp_field

        schema = Schema(self.ds.schema.tecton_schema) if self.ds.schema else None
        arrow_schema = tecton_schema_to_arrow_schema(schema) if schema else None
        if self.spec.timestamp_format and arrow_schema:
            # replace timestamp column type with string,
            # we will convert timestamp with DuckDB (see below)
            timestamp_pos = arrow_schema.names.index(timestamp_field)
            arrow_schema = arrow_schema.set(timestamp_pos, pyarrow.field(timestamp_field, pyarrow.string()))

        proto_format = self.spec.file_format
        if proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_CSV:
            arrow_format = "csv"
        elif proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_JSON:
            arrow_format = "json"
        elif proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_PARQUET:
            arrow_format = "parquet"
        else:
            raise ValueError(batch_data_source_pb2.FileDataSourceFormat.Name(self.spec.file_format))

        fs, path = pyarrow.fs.FileSystem.from_uri(file_uri)
        if isinstance(fs, pyarrow.fs.S3FileSystem):
            options = BotoOfflineStoreOptionsProvider.static_options()
            if options is not None:
                fs = pyarrow.fs.S3FileSystem(
                    access_key=options.access_key_id,
                    secret_key=options.secret_access_key,
                    session_token=options.session_token,
                    # When created via Filesystem.from_uri, the bucket region will be autodetected. This constructor
                    # does not have a bucket from which it can detect the region, so we need to copy it over from the
                    # previous instance.
                    region=fs.region,
                )

        # There seems to be a bug in Arrow related to the explicit schema:
        # when we pass an explicit schema to `dataset` and both resolution and timezone in the timestamp column
        # don't match the schema in parquet files - filters that are pushed down by DuckDB will not work.
        # It is very likely that we will not guess both resolution and timezone correctly.
        # So we won't pass schema for now.
        arrow_schema = arrow_schema if arrow_format != "parquet" else None

        partitioning = None
        partition_filter = None
        # If source supports partitions then only read the relevant partitions
        if (self.start_time or self.end_time) and self.spec.datetime_partition_columns:
            partition_fields = []
            filter_conditions = []
            for i, partition in enumerate(self.spec.datetime_partition_columns):
                partition_col = partition.column_name if partition.column_name else f"_dir_partition_{i}"
                partition_type = None
                partition_value_at_start = None
                partition_value_at_end = None

                if self.start_time:
                    partition_value_at_start, partition_type = _partition_value_and_type_for_time(
                        partition, self.start_time
                    )
                if self.end_time:
                    partition_value_at_end, partition_type = _partition_value_and_type_for_time(
                        partition, self.end_time
                    )

                if partition_value_at_start == partition_value_at_end:
                    # Use the partition path to reduce scanning for metadata when initializing the dataset
                    hive_key = f"{partition.column_name}=" if partition.column_name else ""
                    partition_value = self.start_time.strftime(partition.format_string)
                    path = path.rstrip("/") + f"/{hive_key}{self.start_time.strftime(partition_value)}"
                else:
                    # Otherwise we use a range filter and break so we don't combine hierarchical partition filters
                    partition_fields.append(pyarrow.field(partition_col, partition_type))
                    filter_conditions.append((pyarrow.dataset.field(partition_col) >= partition_value_at_start))
                    filter_conditions.append((pyarrow.dataset.field(partition_col) <= partition_value_at_end))
                    # TODO: combine range filters on hierarchical partitions using nested 'Or' filters
                    break

            # Setup dataset partitioning if we used partition range filters
            if partition_fields:
                partitioning = pyarrow.dataset.partitioning(
                    pyarrow.schema(partition_fields),
                    # default is a directory partition
                    flavor="hive" if self.spec.datetime_partition_columns[0].column_name else None,
                )
                partition_filter = reduce(and_, filter_conditions)

        file_dataset = pyarrow.dataset.dataset(
            source=path, schema=arrow_schema, filesystem=fs, format=arrow_format, partitioning=partitioning
        )
        reader = pyarrow.RecordBatchReader.from_batches(
            file_dataset.schema, file_dataset.to_batches(filter=partition_filter)
        )
        if self.spec.post_processor:

            def _map(input_df: pandas.DataFrame) -> pandas.DataFrame:
                try:
                    return self.spec.post_processor(input_df)
                except Exception as exc:
                    msg = "Post processor function of data source " f"('{self.spec.name}') " f"failed with exception"
                    raise UserCodeError(msg) from exc

            reader = map_batches(reader, _map)

        # ToDo: consider using pyarrow compute instead
        duckdb_session = duckdb_factory.create_connection()
        relation = duckdb_session.from_arrow(reader)
        column_types = dict(zip(relation.columns, relation.dtypes))

        if column_types[timestamp_field] == duckdb.typing.VARCHAR:
            if self.spec.timestamp_format:
                conversion_exp = f"strptime(\"{timestamp_field}\", '{self.spec.timestamp_format}')"
            else:
                conversion_exp = f'CAST("{timestamp_field}" AS TIMESTAMP)'
            relation = relation.select(f'* REPLACE({conversion_exp} AS "{timestamp_field}")')

        if self.start_time:
            if column_types[timestamp_field] == duckdb.typing.TIMESTAMP_TZ:
                start_time = get_timezone_aware_datetime(self.start_time)
            else:
                start_time = self.start_time.replace(tzinfo=None)
            relation = relation.filter(f"\"{timestamp_field}\" >= '{start_time}'")
        if self.end_time:
            if column_types[timestamp_field] == duckdb.typing.TIMESTAMP_TZ:
                end_time = get_timezone_aware_datetime(self.end_time)
            else:
                end_time = self.end_time.replace(tzinfo=None)
            relation = relation.filter(f"\"{timestamp_field}\" < '{end_time}'")

        return relation.fetch_arrow_reader()

    def as_str(self):
        return f"FileDataSourceScanNode for {self.ds.name}"


@attrs.define
class PushTableSourceScanNode(ArrowExecNode):
    ds: specs.DataSourceSpec
    ds_node: Optional[DataSourceNode]
    is_stream: bool = attrs.field()
    start_time: Optional[datetime.datetime] = None
    end_time: Optional[datetime.datetime] = None

    @classmethod
    def from_node_input(cls, query_node: DataSourceScanNode) -> "PushTableSourceScanNode":
        assert isinstance(query_node.ds.batch_source, specs.PushTableSourceSpec)
        return cls.from_node_inputs(query_node, input_node=None)

    @property
    def spec(self) -> specs.PushTableSourceSpec:
        return self.ds.batch_source

    def to_arrow_reader(self, context: ExecutionContext) -> pyarrow.RecordBatchReader:
        from deltalake import DeltaTable

        ds_id = id_helper.IdHelper.from_string(self.ds.id)
        creds = next(
            filter(
                lambda o: o is not None,
                (p.get_s3_options_for_data_source(ds_id) for p in context.offline_store_options_providers),
            ),
            None,
        )
        if not creds:
            msg = f"Unable to retrieve S3 store credentials for data source {self.ds.name}"
            raise Exception(msg)
        storage_options = {
            "AWS_ACCESS_KEY_ID": creds.access_key_id,
            "AWS_SECRET_ACCESS_KEY": creds.secret_access_key,
            "AWS_SESSION_TOKEN": creds.session_token,
            "AWS_S3_LOCKING_PROVIDER": "dynamodb",
            "AWS_REGION": conf.get_or_raise("CLUSTER_REGION"),
        }
        saved_error = None
        for _ in range(20):
            try:
                table = DeltaTable(table_uri=self.spec.ingested_data_location, storage_options=storage_options)
                break
            except OSError as e:
                saved_error = e
                time.sleep(0.1)
        else:
            msg = "Failed to read from S3"
            raise TimeoutError(msg) from saved_error
        ds = table.to_pyarrow_dataset()
        return pyarrow.RecordBatchReader.from_batches(ds.schema, ds.to_batches())

    def as_str(self):
        return f"PushTableSourceScanNode for {self.ds.name}"


def _partition_value_and_type_for_time(
    partition: DatetimePartitionColumnSpec, dt: datetime
) -> Tuple[Union[int, datetime.date], pyarrow.DataType]:
    fmt = partition.format_string
    if fmt == "%-Y" or fmt == "%Y":
        return dt.year, pyarrow.int32()
    elif fmt == "%-m" or fmt == "%m":
        return dt.month, pyarrow.int32()
    elif fmt == "%-d" or fmt == "%d":
        return dt.day, pyarrow.int32()
    elif fmt == "%-H" or fmt == "%H":
        return dt.hour, pyarrow.int32()
    elif fmt == "%Y-%m-%d":
        return dt.date(), pyarrow.date32()
    else:
        msg = f"Datetime format `{fmt}` not supported for partition column {partition.column_name}"
        raise ValueError(msg)


def map_batches(
    input_: pyarrow.RecordBatchReader, map_: Callable[[pandas.DataFrame], pandas.DataFrame]
) -> pyarrow.RecordBatchReader:
    def gen():
        while True:
            try:
                batch = input_.read_next_batch()
            except StopIteration:
                return
            processed = map_(batch.to_pandas())
            yield pyarrow.RecordBatch.from_pandas(processed)

    batches = gen()
    first_batch = next(batches)
    return pyarrow.RecordBatchReader.from_batches(first_batch.schema, itertools.chain([first_batch], batches))



from tecton_proto.args import basic_info__client_pb2 as _basic_info__client_pb2
from tecton_proto.args import data_source__client_pb2 as _data_source__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.common import data_source_type__client_pb2 as _data_source_type__client_pb2
from tecton_proto.common import framework_version__client_pb2 as _framework_version__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import schema_container__client_pb2 as _schema_container__client_pb2
from tecton_proto.common import spark_schema__client_pb2 as _spark_schema__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class VirtualDataSourceArgs(_message.Message):
    __slots__ = ["bigquery_ds_config", "file_ds_config", "forced_batch_schema", "forced_stream_schema", "hive_ds_config", "info", "kafka_ds_config", "kinesis_ds_config", "options", "pandas_batch_config", "prevent_destroy", "push_config", "pyarrow_batch_config", "redshift_ds_config", "schema", "snowflake_ds_config", "spark_batch_config", "spark_stream_config", "type", "unity_ds_config", "version", "virtual_data_source_id"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    BIGQUERY_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    FILE_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    FORCED_BATCH_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    FORCED_STREAM_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    HIVE_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    KAFKA_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    KINESIS_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    PANDAS_BATCH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    PREVENT_DESTROY_FIELD_NUMBER: _ClassVar[int]
    PUSH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    PYARROW_BATCH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    REDSHIFT_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    SNOWFLAKE_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SPARK_BATCH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SPARK_STREAM_CONFIG_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    UNITY_DS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    VIRTUAL_DATA_SOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    bigquery_ds_config: _data_source__client_pb2.BigqueryDataSourceArgs
    file_ds_config: _data_source__client_pb2.FileDataSourceArgs
    forced_batch_schema: _spark_schema__client_pb2.SparkSchema
    forced_stream_schema: _spark_schema__client_pb2.SparkSchema
    hive_ds_config: _data_source__client_pb2.HiveDataSourceArgs
    info: _basic_info__client_pb2.BasicInfo
    kafka_ds_config: _data_source__client_pb2.KafkaDataSourceArgs
    kinesis_ds_config: _data_source__client_pb2.KinesisDataSourceArgs
    options: _containers.ScalarMap[str, str]
    pandas_batch_config: _data_source__client_pb2.PandasBatchConfigArgs
    prevent_destroy: bool
    push_config: _data_source__client_pb2.PushSourceArgs
    pyarrow_batch_config: _data_source__client_pb2.PyArrowBatchConfigArgs
    redshift_ds_config: _data_source__client_pb2.RedshiftDataSourceArgs
    schema: _schema_container__client_pb2.SchemaContainer
    snowflake_ds_config: _data_source__client_pb2.SnowflakeDataSourceArgs
    spark_batch_config: _data_source__client_pb2.SparkBatchConfigArgs
    spark_stream_config: _data_source__client_pb2.SparkStreamConfigArgs
    type: _data_source_type__client_pb2.DataSourceType
    unity_ds_config: _data_source__client_pb2.UnityDataSourceArgs
    version: _framework_version__client_pb2.FrameworkVersion
    virtual_data_source_id: _id__client_pb2.Id
    def __init__(self, virtual_data_source_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., info: _Optional[_Union[_basic_info__client_pb2.BasicInfo, _Mapping]] = ..., version: _Optional[_Union[_framework_version__client_pb2.FrameworkVersion, str]] = ..., prevent_destroy: bool = ..., options: _Optional[_Mapping[str, str]] = ..., hive_ds_config: _Optional[_Union[_data_source__client_pb2.HiveDataSourceArgs, _Mapping]] = ..., file_ds_config: _Optional[_Union[_data_source__client_pb2.FileDataSourceArgs, _Mapping]] = ..., redshift_ds_config: _Optional[_Union[_data_source__client_pb2.RedshiftDataSourceArgs, _Mapping]] = ..., snowflake_ds_config: _Optional[_Union[_data_source__client_pb2.SnowflakeDataSourceArgs, _Mapping]] = ..., spark_batch_config: _Optional[_Union[_data_source__client_pb2.SparkBatchConfigArgs, _Mapping]] = ..., unity_ds_config: _Optional[_Union[_data_source__client_pb2.UnityDataSourceArgs, _Mapping]] = ..., pandas_batch_config: _Optional[_Union[_data_source__client_pb2.PandasBatchConfigArgs, _Mapping]] = ..., bigquery_ds_config: _Optional[_Union[_data_source__client_pb2.BigqueryDataSourceArgs, _Mapping]] = ..., pyarrow_batch_config: _Optional[_Union[_data_source__client_pb2.PyArrowBatchConfigArgs, _Mapping]] = ..., kinesis_ds_config: _Optional[_Union[_data_source__client_pb2.KinesisDataSourceArgs, _Mapping]] = ..., kafka_ds_config: _Optional[_Union[_data_source__client_pb2.KafkaDataSourceArgs, _Mapping]] = ..., spark_stream_config: _Optional[_Union[_data_source__client_pb2.SparkStreamConfigArgs, _Mapping]] = ..., push_config: _Optional[_Union[_data_source__client_pb2.PushSourceArgs, _Mapping]] = ..., schema: _Optional[_Union[_schema_container__client_pb2.SchemaContainer, _Mapping]] = ..., type: _Optional[_Union[_data_source_type__client_pb2.DataSourceType, str]] = ..., forced_batch_schema: _Optional[_Union[_spark_schema__client_pb2.SparkSchema, _Mapping]] = ..., forced_stream_schema: _Optional[_Union[_spark_schema__client_pb2.SparkSchema, _Mapping]] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.common import fco_locator__client_pb2 as _fco_locator__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
FEATURE_VIEW_SIMILARITY_DUPLICATE: FeatureViewSimilarityType
FEATURE_VIEW_SIMILARITY_NEAR_DUPLICATE: FeatureViewSimilarityType
FEATURE_VIEW_SIMILARITY_UNKNOWN: FeatureViewSimilarityType

class FeatureSimilarityAnalysisResult(_message.Message):
    __slots__ = ["analysis_time", "fv_groups"]
    ANALYSIS_TIME_FIELD_NUMBER: _ClassVar[int]
    FV_GROUPS_FIELD_NUMBER: _ClassVar[int]
    analysis_time: _timestamp_pb2.Timestamp
    fv_groups: _containers.RepeatedCompositeFieldContainer[FeatureViewSimilarityGroup]
    def __init__(self, analysis_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., fv_groups: _Optional[_Iterable[_Union[FeatureViewSimilarityGroup, _Mapping]]] = ...) -> None: ...

class FeatureViewSimilarityGroup(_message.Message):
    __slots__ = ["fv_details_first", "fv_locator_first", "pairs", "workspace_state_id_first"]
    FV_DETAILS_FIRST_FIELD_NUMBER: _ClassVar[int]
    FV_LOCATOR_FIRST_FIELD_NUMBER: _ClassVar[int]
    PAIRS_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_STATE_ID_FIRST_FIELD_NUMBER: _ClassVar[int]
    fv_details_first: str
    fv_locator_first: _fco_locator__client_pb2.FcoLocator
    pairs: _containers.RepeatedCompositeFieldContainer[FeatureViewSimilarityPair]
    workspace_state_id_first: _id__client_pb2.Id
    def __init__(self, fv_locator_first: _Optional[_Union[_fco_locator__client_pb2.FcoLocator, _Mapping]] = ..., workspace_state_id_first: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., fv_details_first: _Optional[str] = ..., pairs: _Optional[_Iterable[_Union[FeatureViewSimilarityPair, _Mapping]]] = ...) -> None: ...

class FeatureViewSimilarityPair(_message.Message):
    __slots__ = ["fv_details_second", "fv_locator_second", "similarity", "similarity_description", "type", "workspace_state_id_second"]
    FV_DETAILS_SECOND_FIELD_NUMBER: _ClassVar[int]
    FV_LOCATOR_SECOND_FIELD_NUMBER: _ClassVar[int]
    SIMILARITY_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    SIMILARITY_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_STATE_ID_SECOND_FIELD_NUMBER: _ClassVar[int]
    fv_details_second: str
    fv_locator_second: _fco_locator__client_pb2.FcoLocator
    similarity: float
    similarity_description: str
    type: FeatureViewSimilarityType
    workspace_state_id_second: _id__client_pb2.Id
    def __init__(self, fv_locator_second: _Optional[_Union[_fco_locator__client_pb2.FcoLocator, _Mapping]] = ..., workspace_state_id_second: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., similarity: _Optional[float] = ..., type: _Optional[_Union[FeatureViewSimilarityType, str]] = ..., fv_details_second: _Optional[str] = ..., similarity_description: _Optional[str] = ...) -> None: ...

class FeatureViewSimilarityType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import duration_pb2 as _duration_pb2
from tecton_proto.args import data_source_config__client_pb2 as _data_source_config__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.args import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from tecton_proto.common import secret__client_pb2 as _secret__client_pb2
from tecton_proto.common import spark_schema__client_pb2 as _spark_schema__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
UNITY_CATALOG_ACCESS_MODE_SHARED: UnityCatalogAccessMode
UNITY_CATALOG_ACCESS_MODE_SINGLE_USER: UnityCatalogAccessMode
UNITY_CATALOG_ACCESS_MODE_SINGLE_USER_WITH_FGAC: UnityCatalogAccessMode
UNITY_CATALOG_ACCESS_MODE_UNSPECIFIED: UnityCatalogAccessMode

class BatchDataSourceCommonArgs(_message.Message):
    __slots__ = ["data_delay", "post_processor", "timestamp_field"]
    DATA_DELAY_FIELD_NUMBER: _ClassVar[int]
    POST_PROCESSOR_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_FIELD_NUMBER: _ClassVar[int]
    data_delay: _duration_pb2.Duration
    post_processor: _user_defined_function__client_pb2.UserDefinedFunction
    timestamp_field: str
    def __init__(self, timestamp_field: _Optional[str] = ..., post_processor: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., data_delay: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...) -> None: ...

class BigqueryDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "credentials", "dataset", "location", "project_id", "query", "table"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    CREDENTIALS_FIELD_NUMBER: _ClassVar[int]
    DATASET_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    common_args: BatchDataSourceCommonArgs
    credentials: _secret__client_pb2.SecretReference
    dataset: str
    location: str
    project_id: str
    query: str
    table: str
    def __init__(self, project_id: _Optional[str] = ..., dataset: _Optional[str] = ..., location: _Optional[str] = ..., table: _Optional[str] = ..., query: _Optional[str] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ..., credentials: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ...) -> None: ...

class DatetimePartitionColumnArgs(_message.Message):
    __slots__ = ["column_name", "datepart", "format_string", "zero_padded"]
    COLUMN_NAME_FIELD_NUMBER: _ClassVar[int]
    DATEPART_FIELD_NUMBER: _ClassVar[int]
    FORMAT_STRING_FIELD_NUMBER: _ClassVar[int]
    ZERO_PADDED_FIELD_NUMBER: _ClassVar[int]
    column_name: str
    datepart: str
    format_string: str
    zero_padded: bool
    def __init__(self, column_name: _Optional[str] = ..., datepart: _Optional[str] = ..., zero_padded: bool = ..., format_string: _Optional[str] = ...) -> None: ...

class FileDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "convert_to_glue_format", "datetime_partition_columns", "file_format", "schema_override", "schema_uri", "timestamp_format", "uri"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    CONVERT_TO_GLUE_FORMAT_FIELD_NUMBER: _ClassVar[int]
    DATETIME_PARTITION_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    FILE_FORMAT_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_OVERRIDE_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_URI_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FORMAT_FIELD_NUMBER: _ClassVar[int]
    URI_FIELD_NUMBER: _ClassVar[int]
    common_args: BatchDataSourceCommonArgs
    convert_to_glue_format: bool
    datetime_partition_columns: _containers.RepeatedCompositeFieldContainer[DatetimePartitionColumnArgs]
    file_format: str
    schema_override: _spark_schema__client_pb2.SparkSchema
    schema_uri: str
    timestamp_format: str
    uri: str
    def __init__(self, uri: _Optional[str] = ..., file_format: _Optional[str] = ..., convert_to_glue_format: bool = ..., schema_uri: _Optional[str] = ..., timestamp_format: _Optional[str] = ..., schema_override: _Optional[_Union[_spark_schema__client_pb2.SparkSchema, _Mapping]] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ..., datetime_partition_columns: _Optional[_Iterable[_Union[DatetimePartitionColumnArgs, _Mapping]]] = ...) -> None: ...

class HiveDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "database", "datetime_partition_columns", "table", "timestamp_format"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    DATETIME_PARTITION_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FORMAT_FIELD_NUMBER: _ClassVar[int]
    common_args: BatchDataSourceCommonArgs
    database: str
    datetime_partition_columns: _containers.RepeatedCompositeFieldContainer[DatetimePartitionColumnArgs]
    table: str
    timestamp_format: str
    def __init__(self, table: _Optional[str] = ..., database: _Optional[str] = ..., timestamp_format: _Optional[str] = ..., datetime_partition_columns: _Optional[_Iterable[_Union[DatetimePartitionColumnArgs, _Mapping]]] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ...) -> None: ...

class KafkaDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "kafka_bootstrap_servers", "options", "security_protocol", "ssl_keystore_location", "ssl_keystore_password_secret_id", "ssl_truststore_location", "ssl_truststore_password_secret_id", "topics"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    KAFKA_BOOTSTRAP_SERVERS_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    SECURITY_PROTOCOL_FIELD_NUMBER: _ClassVar[int]
    SSL_KEYSTORE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    SSL_KEYSTORE_PASSWORD_SECRET_ID_FIELD_NUMBER: _ClassVar[int]
    SSL_TRUSTSTORE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    SSL_TRUSTSTORE_PASSWORD_SECRET_ID_FIELD_NUMBER: _ClassVar[int]
    TOPICS_FIELD_NUMBER: _ClassVar[int]
    common_args: StreamDataSourceCommonArgs
    kafka_bootstrap_servers: str
    options: _containers.RepeatedCompositeFieldContainer[Option]
    security_protocol: str
    ssl_keystore_location: str
    ssl_keystore_password_secret_id: str
    ssl_truststore_location: str
    ssl_truststore_password_secret_id: str
    topics: str
    def __init__(self, kafka_bootstrap_servers: _Optional[str] = ..., topics: _Optional[str] = ..., options: _Optional[_Iterable[_Union[Option, _Mapping]]] = ..., ssl_keystore_location: _Optional[str] = ..., ssl_keystore_password_secret_id: _Optional[str] = ..., ssl_truststore_location: _Optional[str] = ..., ssl_truststore_password_secret_id: _Optional[str] = ..., security_protocol: _Optional[str] = ..., common_args: _Optional[_Union[StreamDataSourceCommonArgs, _Mapping]] = ...) -> None: ...

class KinesisDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "initial_stream_position", "options", "region", "stream_name"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    INITIAL_STREAM_POSITION_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    STREAM_NAME_FIELD_NUMBER: _ClassVar[int]
    common_args: StreamDataSourceCommonArgs
    initial_stream_position: _data_source_config__client_pb2.InitialStreamPosition
    options: _containers.RepeatedCompositeFieldContainer[Option]
    region: str
    stream_name: str
    def __init__(self, stream_name: _Optional[str] = ..., region: _Optional[str] = ..., initial_stream_position: _Optional[_Union[_data_source_config__client_pb2.InitialStreamPosition, str]] = ..., options: _Optional[_Iterable[_Union[Option, _Mapping]]] = ..., common_args: _Optional[_Union[StreamDataSourceCommonArgs, _Mapping]] = ...) -> None: ...

class Option(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class PandasBatchConfigArgs(_message.Message):
    __slots__ = ["data_delay", "data_source_function", "secrets", "supports_time_filtering"]
    class SecretsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _secret__client_pb2.SecretReference
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ...) -> None: ...
    DATA_DELAY_FIELD_NUMBER: _ClassVar[int]
    DATA_SOURCE_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    SECRETS_FIELD_NUMBER: _ClassVar[int]
    SUPPORTS_TIME_FILTERING_FIELD_NUMBER: _ClassVar[int]
    data_delay: _duration_pb2.Duration
    data_source_function: _user_defined_function__client_pb2.UserDefinedFunction
    secrets: _containers.MessageMap[str, _secret__client_pb2.SecretReference]
    supports_time_filtering: bool
    def __init__(self, data_source_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., data_delay: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., supports_time_filtering: bool = ..., secrets: _Optional[_Mapping[str, _secret__client_pb2.SecretReference]] = ...) -> None: ...

class PushSourceArgs(_message.Message):
    __slots__ = ["input_schema", "log_offline", "post_processor", "post_processor_mode", "timestamp_field"]
    INPUT_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    LOG_OFFLINE_FIELD_NUMBER: _ClassVar[int]
    POST_PROCESSOR_FIELD_NUMBER: _ClassVar[int]
    POST_PROCESSOR_MODE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_FIELD_NUMBER: _ClassVar[int]
    input_schema: _schema__client_pb2.Schema
    log_offline: bool
    post_processor: _user_defined_function__client_pb2.UserDefinedFunction
    post_processor_mode: _transformation__client_pb2.TransformationMode
    timestamp_field: str
    def __init__(self, log_offline: bool = ..., post_processor: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., input_schema: _Optional[_Union[_schema__client_pb2.Schema, _Mapping]] = ..., post_processor_mode: _Optional[_Union[_transformation__client_pb2.TransformationMode, str]] = ..., timestamp_field: _Optional[str] = ...) -> None: ...

class PyArrowBatchConfigArgs(_message.Message):
    __slots__ = ["data_delay", "data_source_function", "secrets", "supports_time_filtering"]
    class SecretsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _secret__client_pb2.SecretReference
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ...) -> None: ...
    DATA_DELAY_FIELD_NUMBER: _ClassVar[int]
    DATA_SOURCE_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    SECRETS_FIELD_NUMBER: _ClassVar[int]
    SUPPORTS_TIME_FILTERING_FIELD_NUMBER: _ClassVar[int]
    data_delay: _duration_pb2.Duration
    data_source_function: _user_defined_function__client_pb2.UserDefinedFunction
    secrets: _containers.MessageMap[str, _secret__client_pb2.SecretReference]
    supports_time_filtering: bool
    def __init__(self, data_source_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., data_delay: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., supports_time_filtering: bool = ..., secrets: _Optional[_Mapping[str, _secret__client_pb2.SecretReference]] = ...) -> None: ...

class RedshiftDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "endpoint", "query", "table"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    common_args: BatchDataSourceCommonArgs
    endpoint: str
    query: str
    table: str
    def __init__(self, endpoint: _Optional[str] = ..., table: _Optional[str] = ..., query: _Optional[str] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ...) -> None: ...

class SnowflakeDataSourceArgs(_message.Message):
    __slots__ = ["common_args", "database", "password", "query", "role", "schema", "table", "url", "user", "warehouse"]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    WAREHOUSE_FIELD_NUMBER: _ClassVar[int]
    common_args: BatchDataSourceCommonArgs
    database: str
    password: _secret__client_pb2.SecretReference
    query: str
    role: str
    schema: str
    table: str
    url: str
    user: _secret__client_pb2.SecretReference
    warehouse: str
    def __init__(self, url: _Optional[str] = ..., role: _Optional[str] = ..., database: _Optional[str] = ..., schema: _Optional[str] = ..., warehouse: _Optional[str] = ..., table: _Optional[str] = ..., query: _Optional[str] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ..., user: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ..., password: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ...) -> None: ...

class SparkBatchConfigArgs(_message.Message):
    __slots__ = ["data_delay", "data_source_function", "supports_time_filtering"]
    DATA_DELAY_FIELD_NUMBER: _ClassVar[int]
    DATA_SOURCE_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    SUPPORTS_TIME_FILTERING_FIELD_NUMBER: _ClassVar[int]
    data_delay: _duration_pb2.Duration
    data_source_function: _user_defined_function__client_pb2.UserDefinedFunction
    supports_time_filtering: bool
    def __init__(self, data_source_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., data_delay: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., supports_time_filtering: bool = ...) -> None: ...

class SparkStreamConfigArgs(_message.Message):
    __slots__ = ["data_source_function"]
    DATA_SOURCE_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    data_source_function: _user_defined_function__client_pb2.UserDefinedFunction
    def __init__(self, data_source_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ...) -> None: ...

class StreamDataSourceCommonArgs(_message.Message):
    __slots__ = ["deduplication_columns", "post_processor", "timestamp_field", "watermark_delay_threshold"]
    DEDUPLICATION_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    POST_PROCESSOR_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_FIELD_NUMBER: _ClassVar[int]
    WATERMARK_DELAY_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    deduplication_columns: _containers.RepeatedScalarFieldContainer[str]
    post_processor: _user_defined_function__client_pb2.UserDefinedFunction
    timestamp_field: str
    watermark_delay_threshold: _duration_pb2.Duration
    def __init__(self, timestamp_field: _Optional[str] = ..., watermark_delay_threshold: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., post_processor: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., deduplication_columns: _Optional[_Iterable[str]] = ...) -> None: ...

class UnityDataSourceArgs(_message.Message):
    __slots__ = ["access_mode", "catalog", "common_args", "datetime_partition_columns", "schema", "table", "timestamp_format"]
    ACCESS_MODE_FIELD_NUMBER: _ClassVar[int]
    CATALOG_FIELD_NUMBER: _ClassVar[int]
    COMMON_ARGS_FIELD_NUMBER: _ClassVar[int]
    DATETIME_PARTITION_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FORMAT_FIELD_NUMBER: _ClassVar[int]
    access_mode: UnityCatalogAccessMode
    catalog: str
    common_args: BatchDataSourceCommonArgs
    datetime_partition_columns: _containers.RepeatedCompositeFieldContainer[DatetimePartitionColumnArgs]
    schema: str
    table: str
    timestamp_format: str
    def __init__(self, catalog: _Optional[str] = ..., schema: _Optional[str] = ..., table: _Optional[str] = ..., common_args: _Optional[_Union[BatchDataSourceCommonArgs, _Mapping]] = ..., timestamp_format: _Optional[str] = ..., datetime_partition_columns: _Optional[_Iterable[_Union[DatetimePartitionColumnArgs, _Mapping]]] = ..., access_mode: _Optional[_Union[UnityCatalogAccessMode, str]] = ...) -> None: ...

class UnityCatalogAccessMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.common import container_image__client_pb2 as _container_image__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import server_group_status__client_pb2 as _server_group_status__client_pb2
from tecton_proto.common import server_group_type__client_pb2 as _server_group_type__client_pb2
from tecton_proto.realtime import instance_group__client_pb2 as _instance_group__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
SCALING_POLICY_TYPE_TARGET_TRACKING: ScalingPolicyType
SCALING_POLICY_TYPE_UNSPECIFIED: ScalingPolicyType

class AutoscalingPolicy(_message.Message):
    __slots__ = ["autoscaling_enabled", "name", "target_tracking_policy"]
    AUTOSCALING_ENABLED_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TARGET_TRACKING_POLICY_FIELD_NUMBER: _ClassVar[int]
    autoscaling_enabled: bool
    name: str
    target_tracking_policy: TargetTrackingPolicy
    def __init__(self, name: _Optional[str] = ..., autoscaling_enabled: bool = ..., target_tracking_policy: _Optional[_Union[TargetTrackingPolicy, _Mapping]] = ...) -> None: ...

class CustomMetric(_message.Message):
    __slots__ = ["name", "statistic", "unit"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATISTIC_FIELD_NUMBER: _ClassVar[int]
    UNIT_FIELD_NUMBER: _ClassVar[int]
    name: str
    statistic: str
    unit: str
    def __init__(self, name: _Optional[str] = ..., statistic: _Optional[str] = ..., unit: _Optional[str] = ...) -> None: ...

class FeatureServerGroupAWSResources(_message.Message):
    __slots__ = ["autoscaling_group_id", "launch_template_active_version", "launch_template_id", "launch_template_pending_version", "loadbalancer_listener_rule_arn", "target_group_arn"]
    AUTOSCALING_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    LAUNCH_TEMPLATE_ACTIVE_VERSION_FIELD_NUMBER: _ClassVar[int]
    LAUNCH_TEMPLATE_ID_FIELD_NUMBER: _ClassVar[int]
    LAUNCH_TEMPLATE_PENDING_VERSION_FIELD_NUMBER: _ClassVar[int]
    LOADBALANCER_LISTENER_RULE_ARN_FIELD_NUMBER: _ClassVar[int]
    TARGET_GROUP_ARN_FIELD_NUMBER: _ClassVar[int]
    autoscaling_group_id: str
    launch_template_active_version: str
    launch_template_id: str
    launch_template_pending_version: str
    loadbalancer_listener_rule_arn: str
    target_group_arn: str
    def __init__(self, loadbalancer_listener_rule_arn: _Optional[str] = ..., autoscaling_group_id: _Optional[str] = ..., target_group_arn: _Optional[str] = ..., launch_template_id: _Optional[str] = ..., launch_template_active_version: _Optional[str] = ..., launch_template_pending_version: _Optional[str] = ...) -> None: ...

class FeatureServerGroupConfig(_message.Message):
    __slots__ = ["autoscaling_enabled", "desired_instances", "docker_image_fqn", "environment", "http_health_check_path", "http_port", "instance_type", "max_instances", "metrics_port", "min_instances"]
    class EnvironmentEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    AUTOSCALING_ENABLED_FIELD_NUMBER: _ClassVar[int]
    DESIRED_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    DOCKER_IMAGE_FQN_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    HTTP_HEALTH_CHECK_PATH_FIELD_NUMBER: _ClassVar[int]
    HTTP_PORT_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    MAX_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    METRICS_PORT_FIELD_NUMBER: _ClassVar[int]
    MIN_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    autoscaling_enabled: bool
    desired_instances: int
    docker_image_fqn: str
    environment: _containers.ScalarMap[str, str]
    http_health_check_path: str
    http_port: int
    instance_type: str
    max_instances: int
    metrics_port: int
    min_instances: int
    def __init__(self, instance_type: _Optional[str] = ..., docker_image_fqn: _Optional[str] = ..., http_port: _Optional[int] = ..., http_health_check_path: _Optional[str] = ..., metrics_port: _Optional[int] = ..., autoscaling_enabled: bool = ..., min_instances: _Optional[int] = ..., max_instances: _Optional[int] = ..., desired_instances: _Optional[int] = ..., environment: _Optional[_Mapping[str, str]] = ...) -> None: ...

class FeatureServerGroupState(_message.Message):
    __slots__ = ["active_config", "aws_resources", "pending_config"]
    ACTIVE_CONFIG_FIELD_NUMBER: _ClassVar[int]
    AWS_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    PENDING_CONFIG_FIELD_NUMBER: _ClassVar[int]
    active_config: FeatureServerGroupConfig
    aws_resources: FeatureServerGroupAWSResources
    pending_config: FeatureServerGroupConfig
    def __init__(self, active_config: _Optional[_Union[FeatureServerGroupConfig, _Mapping]] = ..., pending_config: _Optional[_Union[FeatureServerGroupConfig, _Mapping]] = ..., aws_resources: _Optional[_Union[FeatureServerGroupAWSResources, _Mapping]] = ...) -> None: ...

class PredefinedMetric(_message.Message):
    __slots__ = ["type"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: str
    def __init__(self, type: _Optional[str] = ...) -> None: ...

class ServerGroupState(_message.Message):
    __slots__ = ["autoscaling_policy", "created_at", "desired_nodes", "feature_server_group_state", "last_updated_at", "max_nodes", "min_nodes", "name", "server_group_id", "server_group_state_id", "status", "status_details", "transform_server_group_state", "type", "workspace", "workspace_state_id"]
    AUTOSCALING_POLICY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    DESIRED_NODES_FIELD_NUMBER: _ClassVar[int]
    FEATURE_SERVER_GROUP_STATE_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    MAX_NODES_FIELD_NUMBER: _ClassVar[int]
    MIN_NODES_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_STATE_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_DETAILS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TRANSFORM_SERVER_GROUP_STATE_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_STATE_ID_FIELD_NUMBER: _ClassVar[int]
    autoscaling_policy: AutoscalingPolicy
    created_at: _timestamp_pb2.Timestamp
    desired_nodes: int
    feature_server_group_state: FeatureServerGroupState
    last_updated_at: _timestamp_pb2.Timestamp
    max_nodes: int
    min_nodes: int
    name: str
    server_group_id: _id__client_pb2.Id
    server_group_state_id: _id__client_pb2.Id
    status: _server_group_status__client_pb2.ServerGroupStatus
    status_details: str
    transform_server_group_state: TransformServerGroupState
    type: _server_group_type__client_pb2.ServerGroupType
    workspace: str
    workspace_state_id: _id__client_pb2.Id
    def __init__(self, server_group_state_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., server_group_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., name: _Optional[str] = ..., status: _Optional[_Union[_server_group_status__client_pb2.ServerGroupStatus, str]] = ..., type: _Optional[_Union[_server_group_type__client_pb2.ServerGroupType, str]] = ..., autoscaling_policy: _Optional[_Union[AutoscalingPolicy, _Mapping]] = ..., status_details: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., last_updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., min_nodes: _Optional[int] = ..., max_nodes: _Optional[int] = ..., desired_nodes: _Optional[int] = ..., transform_server_group_state: _Optional[_Union[TransformServerGroupState, _Mapping]] = ..., feature_server_group_state: _Optional[_Union[FeatureServerGroupState, _Mapping]] = ..., workspace: _Optional[str] = ..., workspace_state_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ...) -> None: ...

class TargetTrackingPolicy(_message.Message):
    __slots__ = ["custom_metric", "predefined_metric", "scale_in_disabled", "target_value"]
    CUSTOM_METRIC_FIELD_NUMBER: _ClassVar[int]
    PREDEFINED_METRIC_FIELD_NUMBER: _ClassVar[int]
    SCALE_IN_DISABLED_FIELD_NUMBER: _ClassVar[int]
    TARGET_VALUE_FIELD_NUMBER: _ClassVar[int]
    custom_metric: CustomMetric
    predefined_metric: PredefinedMetric
    scale_in_disabled: bool
    target_value: float
    def __init__(self, target_value: _Optional[float] = ..., scale_in_disabled: bool = ..., predefined_metric: _Optional[_Union[PredefinedMetric, _Mapping]] = ..., custom_metric: _Optional[_Union[CustomMetric, _Mapping]] = ...) -> None: ...

class TransformServerGroupState(_message.Message):
    __slots__ = ["autoscaling_policy", "environment_id", "environment_name", "environment_variables", "image_info", "instance_group_handle", "secrets_config"]
    class EnvironmentVariablesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    AUTOSCALING_POLICY_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: _ClassVar[int]
    IMAGE_INFO_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_GROUP_HANDLE_FIELD_NUMBER: _ClassVar[int]
    SECRETS_CONFIG_FIELD_NUMBER: _ClassVar[int]
    autoscaling_policy: AutoscalingPolicy
    environment_id: str
    environment_name: str
    environment_variables: _containers.ScalarMap[str, str]
    image_info: _container_image__client_pb2.ContainerImage
    instance_group_handle: _instance_group__client_pb2.InstanceGroupHandle
    secrets_config: TransformServerSecretsConfig
    def __init__(self, environment_id: _Optional[str] = ..., environment_name: _Optional[str] = ..., image_info: _Optional[_Union[_container_image__client_pb2.ContainerImage, _Mapping]] = ..., environment_variables: _Optional[_Mapping[str, str]] = ..., autoscaling_policy: _Optional[_Union[AutoscalingPolicy, _Mapping]] = ..., instance_group_handle: _Optional[_Union[_instance_group__client_pb2.InstanceGroupHandle, _Mapping]] = ..., secrets_config: _Optional[_Union[TransformServerSecretsConfig, _Mapping]] = ...) -> None: ...

class TransformServerSecretsConfig(_message.Message):
    __slots__ = ["last_updated", "service_account_id", "service_account_key_secret_name"]
    LAST_UPDATED_FIELD_NUMBER: _ClassVar[int]
    SERVICE_ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    SERVICE_ACCOUNT_KEY_SECRET_NAME_FIELD_NUMBER: _ClassVar[int]
    last_updated: _timestamp_pb2.Timestamp
    service_account_id: str
    service_account_key_secret_name: str
    def __init__(self, service_account_id: _Optional[str] = ..., service_account_key_secret_name: _Optional[str] = ..., last_updated: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ScalingPolicyType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from abc import ABC
from typing import List
from typing import Optional
from typing import Union

from tecton_core.data_types import DataType
from tecton_core.data_types import data_type_from_proto
from tecton_core.errors import TectonInternalError
from tecton_core.specs import utils
from tecton_proto.common import calculation_node__client_pb2 as calculation_node_pb2


@utils.frozen_strict
class AbstractSyntaxTreeNodeSpec(ABC):
    dtype: Optional[DataType]

    @classmethod
    def from_proto(cls, node: calculation_node_pb2.AbstractSyntaxTreeNode) -> "AbstractSyntaxTreeNodeSpec":
        value_type = node.WhichOneof("value")
        if value_type == "literal_value":
            return LiteralValueNodeSpec.from_proto(node)
        elif value_type == "column_reference":
            return ColumnReferenceNodeSpec.from_proto(node)
        elif value_type == "operation":
            return OperationNodeSpec.from_proto(node)
        elif value_type == "date_part":
            return DatePartNodeSpec.from_proto(node)
        elif value_type == "case_statement":
            return CaseStatementNodeSpec.from_proto(node)
        else:
            msg = f"Unknown AbstractSyntaxTreeNode Type {value_type}"
            raise TectonInternalError(msg)


@utils.frozen_strict
class OperationNodeSpec(AbstractSyntaxTreeNodeSpec):
    dtype: DataType
    operation: calculation_node_pb2.OperationType
    operands: List[AbstractSyntaxTreeNodeSpec]

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.AbstractSyntaxTreeNode) -> "OperationNodeSpec":
        operation_proto = proto.operation
        return cls(
            dtype=data_type_from_proto(proto.dtype),
            operation=operation_proto.operation,
            operands=[AbstractSyntaxTreeNodeSpec.from_proto(operand) for operand in operation_proto.operands],
        )


@utils.frozen_strict
class LiteralValueNodeSpec(AbstractSyntaxTreeNodeSpec):
    dtype: Optional[DataType]
    value: Optional[Union[float, int, bool, str]]

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.AbstractSyntaxTreeNode) -> "LiteralValueNodeSpec":
        literal_value_proto = proto.literal_value
        value_type = literal_value_proto.WhichOneof("value")
        if value_type == "null_value":
            return cls(dtype=None, value=None)
        else:
            return cls(dtype=data_type_from_proto(proto.dtype), value=getattr(literal_value_proto, value_type))


@utils.frozen_strict
class ColumnReferenceNodeSpec(AbstractSyntaxTreeNodeSpec):
    dtype: DataType
    value: str

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.AbstractSyntaxTreeNode) -> "ColumnReferenceNodeSpec":
        return cls(dtype=data_type_from_proto(proto.dtype), value=proto.column_reference)


@utils.frozen_strict
class DatePartNodeSpec(AbstractSyntaxTreeNodeSpec):
    value: calculation_node_pb2.DatePart.ValueType

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.AbstractSyntaxTreeNode) -> "DatePartNodeSpec":
        return cls(value=proto.date_part, dtype=None)


@utils.frozen_strict
class WhenClauseSpec:
    condition: AbstractSyntaxTreeNodeSpec
    result: AbstractSyntaxTreeNodeSpec

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.WhenClause) -> "WhenClauseSpec":
        return cls(
            condition=AbstractSyntaxTreeNodeSpec.from_proto(proto.condition),
            result=AbstractSyntaxTreeNodeSpec.from_proto(proto.result),
        )


@utils.frozen_strict
class ElseClauseSpec:
    result: AbstractSyntaxTreeNodeSpec

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.ElseClause) -> "ElseClauseSpec":
        if proto.HasField("result"):
            return cls(result=AbstractSyntaxTreeNodeSpec.from_proto(proto.result))


@utils.frozen_strict
class CaseStatementNodeSpec(AbstractSyntaxTreeNodeSpec):
    dtype: DataType
    when_clauses: List[WhenClauseSpec]
    else_clause: Optional[ElseClauseSpec] = None

    @classmethod
    def from_proto(cls, proto: calculation_node_pb2.AbstractSyntaxTreeNode) -> "CaseStatementNodeSpec":
        return cls(
            data_type_from_proto(proto.dtype),
            when_clauses=[
                WhenClauseSpec.from_proto(when_clause_proto) for when_clause_proto in proto.case_statement.when_clauses
            ],
            else_clause=ElseClauseSpec.from_proto(proto.case_statement.else_clause),
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/serviceaccounts/service_accounts_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auditlog import metadata__client_pb2 as tecton__proto_dot_auditlog_dot_metadata____client__pb2
from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2
from tecton_proto.data import service_account__client_pb2 as tecton__proto_dot_data_dot_service__account____client__pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as tecton__proto_dot_metadataservice_dot_metadata__service____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nCtecton_proto/serviceaccounts/service_accounts_service__client.proto\x12\x1ctecton_proto.serviceaccounts\x1a,tecton_proto/auditlog/metadata__client.proto\x1a\'tecton_proto/auth/service__client.proto\x1a/tecton_proto/data/service_account__client.proto\x1a;tecton_proto/metadataservice/metadata_service__client.proto\"<\n!CreateServiceAccountSecretRequest\x12\x17\n\x02id\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x02id\"`\n\"CreateServiceAccountSecretResponse\x12:\n\x06secret\x18\x01 \x01(\x0b\x32\".tecton_proto.data.NewClientSecretR\x06secret\"2\n ListServiceAccountSecretsRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\"q\n!ListServiceAccountSecretsResponse\x12L\n\x0e\x63lient_secrets\x18\x01 \x03(\x0b\x32%.tecton_proto.data.MaskedClientSecretR\rclientSecrets\"\x82\x01\n#ActivateServiceAccountSecretRequest\x12\x35\n\x12service_account_id\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x10serviceAccountId\x12$\n\tsecret_id\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x08secretId\"e\n$ActivateServiceAccountSecretResponse\x12=\n\x06secret\x18\x01 \x01(\x0b\x32%.tecton_proto.data.MaskedClientSecretR\x06secret\"\x84\x01\n%DeactivateServiceAccountSecretRequest\x12\x35\n\x12service_account_id\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x10serviceAccountId\x12$\n\tsecret_id\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x08secretId\"g\n&DeactivateServiceAccountSecretResponse\x12=\n\x06secret\x18\x01 \x01(\x0b\x32%.tecton_proto.data.MaskedClientSecretR\x06secret\"\x80\x01\n!DeleteServiceAccountSecretRequest\x12\x35\n\x12service_account_id\x18\x01 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x10serviceAccountId\x12$\n\tsecret_id\x18\x02 \x01(\tB\x07\xea\x99\xbd\x46\x02\x08\x02R\x08secretId\"$\n\"DeleteServiceAccountSecretResponse2\xcd\x11\n\x16ServiceAccountsService\x12\xe9\x01\n\x1a\x43reateSingleServiceAccount\x12\x39.tecton_proto.metadataservice.CreateServiceAccountRequest\x1a:.tecton_proto.metadataservice.CreateServiceAccountResponse\"T\xa2\xbc\xe6\xc0\x05,**tecton/organization#create_service_account\xf2\x99\xbd\x46\x1d\x08\x01\x12\x16\x63reate_service_account\x1a\x01\x31\x12\x94\x01\n\x15SearchServiceAccounts\x12\x37.tecton_proto.metadataservice.GetServiceAccountsRequest\x1a\x38.tecton_proto.metadataservice.GetServiceAccountsResponse\"\x08\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xf4\x01\n\x1aUpdateSingleServiceAccount\x12\x39.tecton_proto.metadataservice.UpdateServiceAccountRequest\x1a:.tecton_proto.metadataservice.UpdateServiceAccountResponse\"_\xa2\xbc\xe6\xc0\x05\x37\x1a\x06\x08\x02\x12\x02id*-tecton/service_account#manage_service_account\xf2\x99\xbd\x46\x1d\x08\x01\x12\x16update_service_account\x1a\x01\x31\x12\xf4\x01\n\x1a\x44\x65leteSingleServiceAccount\x12\x39.tecton_proto.metadataservice.DeleteServiceAccountRequest\x1a:.tecton_proto.metadataservice.DeleteServiceAccountResponse\"_\xa2\xbc\xe6\xc0\x05\x37\x1a\x06\x08\x02\x12\x02id*-tecton/service_account#manage_service_account\xf2\x99\xbd\x46\x1d\x08\x01\x12\x16\x64\x65lete_service_account\x1a\x01\x31\x12\x85\x02\n\x1a\x43reateServiceAccountSecret\x12?.tecton_proto.serviceaccounts.CreateServiceAccountSecretRequest\x1a@.tecton_proto.serviceaccounts.CreateServiceAccountSecretResponse\"d\xa2\xbc\xe6\xc0\x05\x35\x1a\x06\x08\x02\x12\x02id*+tecton/service_account#manage_client_secret\xf2\x99\xbd\x46$\x08\x01\x12\x1d\x63reate_service_account_secret\x1a\x01\x31\x12\xd9\x01\n\x19ListServiceAccountSecrets\x12>.tecton_proto.serviceaccounts.ListServiceAccountSecretsRequest\x1a?.tecton_proto.serviceaccounts.ListServiceAccountSecretsResponse\";\xa2\xbc\xe6\xc0\x05\x35\x1a\x06\x08\x02\x12\x02id*+tecton/service_account#manage_client_secret\x12\x9d\x02\n\x1c\x41\x63tivateServiceAccountSecret\x12\x41.tecton_proto.serviceaccounts.ActivateServiceAccountSecretRequest\x1a\x42.tecton_proto.serviceaccounts.ActivateServiceAccountSecretResponse\"v\xa2\xbc\xe6\xc0\x05\x45\x1a\x16\x08\x02\x12\x12service_account_id*+tecton/service_account#manage_client_secret\xf2\x99\xbd\x46&\x08\x01\x12\x1f\x61\x63tivate_service_account_secret\x1a\x01\x31\x12\xa5\x02\n\x1e\x44\x65\x61\x63tivateServiceAccountSecret\x12\x43.tecton_proto.serviceaccounts.DeactivateServiceAccountSecretRequest\x1a\x44.tecton_proto.serviceaccounts.DeactivateServiceAccountSecretResponse\"x\xa2\xbc\xe6\xc0\x05\x45\x1a\x16\x08\x02\x12\x12service_account_id*+tecton/service_account#manage_client_secret\xf2\x99\xbd\x46(\x08\x01\x12!deactivate_service_account_secret\x1a\x01\x31\x12\x95\x02\n\x1a\x44\x65leteServiceAccountSecret\x12?.tecton_proto.serviceaccounts.DeleteServiceAccountSecretRequest\x1a@.tecton_proto.serviceaccounts.DeleteServiceAccountSecretResponse\"t\xa2\xbc\xe6\xc0\x05\x45\x1a\x16\x08\x02\x12\x12service_account_id*+tecton/service_account#manage_client_secret\xf2\x99\xbd\x46$\x08\x01\x12\x1d\x64\x65lete_service_account_secret\x1a\x01\x31\x42n\n\x1a\x63om.tecton.serviceaccountsB\x1bServiceAccountsServiceProtoP\x01Z1github.com/tecton-ai/tecton_proto/serviceaccounts')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.serviceaccounts.service_accounts_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.serviceaccountsB\033ServiceAccountsServiceProtoP\001Z1github.com/tecton-ai/tecton_proto/serviceaccounts'
  _CREATESERVICEACCOUNTSECRETREQUEST.fields_by_name['id']._options = None
  _CREATESERVICEACCOUNTSECRETREQUEST.fields_by_name['id']._serialized_options = b'\352\231\275F\002\010\002'
  _ACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._options = None
  _ACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._serialized_options = b'\352\231\275F\002\010\002'
  _ACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._options = None
  _ACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._serialized_options = b'\352\231\275F\002\010\002'
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._options = None
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._serialized_options = b'\352\231\275F\002\010\002'
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._options = None
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._serialized_options = b'\352\231\275F\002\010\002'
  _DELETESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._options = None
  _DELETESERVICEACCOUNTSECRETREQUEST.fields_by_name['service_account_id']._serialized_options = b'\352\231\275F\002\010\002'
  _DELETESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._options = None
  _DELETESERVICEACCOUNTSECRETREQUEST.fields_by_name['secret_id']._serialized_options = b'\352\231\275F\002\010\002'
  _SERVICEACCOUNTSSERVICE.methods_by_name['CreateSingleServiceAccount']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['CreateSingleServiceAccount']._serialized_options = b'\242\274\346\300\005,**tecton/organization#create_service_account\362\231\275F\035\010\001\022\026create_service_account\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['SearchServiceAccounts']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['SearchServiceAccounts']._serialized_options = b'\242\274\346\300\005\002@\001'
  _SERVICEACCOUNTSSERVICE.methods_by_name['UpdateSingleServiceAccount']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['UpdateSingleServiceAccount']._serialized_options = b'\242\274\346\300\0057\032\006\010\002\022\002id*-tecton/service_account#manage_service_account\362\231\275F\035\010\001\022\026update_service_account\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeleteSingleServiceAccount']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeleteSingleServiceAccount']._serialized_options = b'\242\274\346\300\0057\032\006\010\002\022\002id*-tecton/service_account#manage_service_account\362\231\275F\035\010\001\022\026delete_service_account\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['CreateServiceAccountSecret']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['CreateServiceAccountSecret']._serialized_options = b'\242\274\346\300\0055\032\006\010\002\022\002id*+tecton/service_account#manage_client_secret\362\231\275F$\010\001\022\035create_service_account_secret\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['ListServiceAccountSecrets']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['ListServiceAccountSecrets']._serialized_options = b'\242\274\346\300\0055\032\006\010\002\022\002id*+tecton/service_account#manage_client_secret'
  _SERVICEACCOUNTSSERVICE.methods_by_name['ActivateServiceAccountSecret']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['ActivateServiceAccountSecret']._serialized_options = b'\242\274\346\300\005E\032\026\010\002\022\022service_account_id*+tecton/service_account#manage_client_secret\362\231\275F&\010\001\022\037activate_service_account_secret\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeactivateServiceAccountSecret']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeactivateServiceAccountSecret']._serialized_options = b'\242\274\346\300\005E\032\026\010\002\022\022service_account_id*+tecton/service_account#manage_client_secret\362\231\275F(\010\001\022!deactivate_service_account_secret\032\0011'
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeleteServiceAccountSecret']._options = None
  _SERVICEACCOUNTSSERVICE.methods_by_name['DeleteServiceAccountSecret']._serialized_options = b'\242\274\346\300\005E\032\026\010\002\022\022service_account_id*+tecton/service_account#manage_client_secret\362\231\275F$\010\001\022\035delete_service_account_secret\032\0011'
  _CREATESERVICEACCOUNTSECRETREQUEST._serialized_start=298
  _CREATESERVICEACCOUNTSECRETREQUEST._serialized_end=358
  _CREATESERVICEACCOUNTSECRETRESPONSE._serialized_start=360
  _CREATESERVICEACCOUNTSECRETRESPONSE._serialized_end=456
  _LISTSERVICEACCOUNTSECRETSREQUEST._serialized_start=458
  _LISTSERVICEACCOUNTSECRETSREQUEST._serialized_end=508
  _LISTSERVICEACCOUNTSECRETSRESPONSE._serialized_start=510
  _LISTSERVICEACCOUNTSECRETSRESPONSE._serialized_end=623
  _ACTIVATESERVICEACCOUNTSECRETREQUEST._serialized_start=626
  _ACTIVATESERVICEACCOUNTSECRETREQUEST._serialized_end=756
  _ACTIVATESERVICEACCOUNTSECRETRESPONSE._serialized_start=758
  _ACTIVATESERVICEACCOUNTSECRETRESPONSE._serialized_end=859
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST._serialized_start=862
  _DEACTIVATESERVICEACCOUNTSECRETREQUEST._serialized_end=994
  _DEACTIVATESERVICEACCOUNTSECRETRESPONSE._serialized_start=996
  _DEACTIVATESERVICEACCOUNTSECRETRESPONSE._serialized_end=1099
  _DELETESERVICEACCOUNTSECRETREQUEST._serialized_start=1102
  _DELETESERVICEACCOUNTSECRETREQUEST._serialized_end=1230
  _DELETESERVICEACCOUNTSECRETRESPONSE._serialized_start=1232
  _DELETESERVICEACCOUNTSECRETRESPONSE._serialized_end=1268
  _SERVICEACCOUNTSSERVICE._serialized_start=1271
  _SERVICEACCOUNTSSERVICE._serialized_end=3524
# @@protoc_insertion_point(module_scope)
